/*
 * xgraph - A Simple Plotter for tk
 *
 * David Harrison
 * University of California,  Berkeley
 * 1986, 1987, 1988, 1989
 *
 * Please see copyright.h concerning the formal reproduction rights
 * of this software.
 *
 * 14-Dec-92  Mark Maimone (mwm@cmu.edu) -- Make Postscript the default
 *	device, read PRINTER environment variable. (thanks to Erik Riedel
 *	er1p+@cmu.edu for finding the Postscript default)
 *
 * 5-Nov-92  Mark Maimone (mwm@cmu.edu) -- Allow "-" to mean stdin on the
 *      command line.
 *
 * 27-Oct-92  Mark Maimone (mwm@cmu.edu) -- extended datafile format to
 * 	accept "Y" as well as "X Y" input.
 *
 * Jun-97 Ben Allan ported to tcl/tk.
 */

#include <stdio.h>
#include <assert.h>
#include <math.h>
#include <limits.h>
/* #include <pwd.h> */
#include <ctype.h>
#include <sys/types.h> /* for caddr_t */
#include "tcl.h"
#include "tk.h"
#include "tkxg.h"
#include "xgout.h"
#include "nxgraph.h"
#include "params.h"

#define ZOOM
#define TOOLBOX
#define NEW_READER

#ifndef FLT_MAX
#ifndef MAXFLOAT
#define MAXFLOAT 1E+37
#endif
#else 
#define MAXFLOAT FLT_MAX
#endif

#define BIGINT		0xfffffff

#define GRIDPOWER 	10
#define INITSIZE 	128

#define CONTROL_D	'\004'
#define CONTROL_C	'\003'
#define TILDE		'~'

#define NORMSIZE	600
#define MINDIM		100

#define BTNPAD		1
#define BTNINTER	3

#ifndef NULL
#define NULL ((char *)0)
#endif

#define MAX(a,b)	((a) > (b) ? (a) : (b))
#define MIN(a,b)	((a) < (b) ? (a) : (b))
#define ABS(x)		((x) < 0 ? -(x) : (x))
#define ZERO_THRES	1.0E-07

/* To get around an inaccurate log */
#define nlog10(x)	(x == 0.0 ? 0.0 : log10(x) + 1e-15)

#define ISCOLOR		(wi->dev_info.dev_flags & D_COLOR)

#define PIXVALUE(set) 	((set) % MAXATTR)

#define LINESTYLE(set) \
(ISCOLOR ?  ((set)/MAXATTR) : ((set) % MAXATTR))

#define MARKSTYLE(set) \
(colorMark ? COLMARK(set) : BWMARK(set))

#define COLMARK(set) \
((set) / MAXATTR)

#define BWMARK(set) \
((set) % MAXATTR)

#define LOG_X	0x01
#define LOG_Y	0x02


/*
 * Default settings for xgraph parameters
 */

#define DEF_BORDER_WIDTH	"2"
#define DEF_BORDER_COLOR	"Black"
#define DEF_TITLE_TEXT		"X Graph"
#define DEF_XUNIT_TEXT		"X"
#define DEF_YUNIT_TEXT		"Y"
#define DEF_TICK_FLAG		"0"
#define DEF_MARK_FLAG		"0"
#define DEF_PIXMARK_FLAG	"0"
#define DEF_LARGEPIX_FLAG	"0"
#define DEF_DIFFMARK_FLAG	"0"
#define DEF_BB_FLAG		"0"
#define DEF_NOLINE_FLAG		"0"
#define DEF_LOGX_FLAG		"0"
#define DEF_LOGY_FLAG		"0"
#define DEF_NEGATEX_FLAG	"0"
#define DEF_NEGATEY_FLAG	"0"
#define DEF_BAR_FLAG		"0"
#define DEF_BAR_BASE		"0.0"
#define DEF_BAR_WIDTH		"-1.0"
#define DEF_LINE_WIDTH		"0"
#define DEF_GRID_SIZE		"0"
#define DEF_GRID_STYLE		"10"
#define DEF_LABEL_FONT		"-*-helvetica-*-r-*-*-12-*"
#define DEF_TITLE_FONT		"-*-helvetica-*-r-*-*-18-*"
#define DEF_GEOMETRY		""
#define DEF_WIDTH		"600"
#define DEF_HEIGHT		"600"
#define DEF_REVERSE		"0"
#define DEF_DEVICE		"Postscript"
#define DEF_DISPOSITION		"To Device"
#define DEF_FILEORDEV		""

#define DEF_MARKER_FLAG		"0"
#define DEF_DIFFMARK_FLAG	"0"
#define DEF_PIXMARK_FLAG	"0"
#define DEF_LARGEPIX_FLAG	"0"

/* Low > High means set it based on the data */
#define DEF_LOW_LIMIT		"1.0"
#define DEF_HIGH_LIMIT		"0.0"

/* Black and white defaults */
#define DEF_BW_BACKGROUND	"white"
#define DEF_BW_BORDER		"black"
#define DEF_BW_ZEROCOLOR	"black"
#define DEF_BW_ZEROWIDTH	"3"
#define DEF_BW_ZEROSTYLE	"1"
#define DEF_BW_FOREGROUND	"black"

/* Color defaults */
#define DEF_COL_BACKGROUND	"#ccc"
#define DEF_COL_BORDER		"black"
#define DEF_COL_ZEROCOLOR	"white"
#define DEF_COL_ZEROWIDTH	"0"
#define DEF_COL_ZEROSTYLE	"1"
#define DEF_COL_FOREGROUND	"black"
#define DEF_COL_FIRSTSTYLE	"1"

/* widget id counter and base name. returns from create will be 
 * TKBASENAMEname_counter
 */
static unsigned int name_counter = 0;
#define TKBASENAME "xgraph"

/* Default line styles */
static char *defStyle[MAXATTR] = {
    "1", "10", "11110000", "010111", "1110",
    "1111111100000000", "11001111", "0011000111"
};

/* Default color names */
static char *defColors[MAXATTR] = {
    "red", "SpringGreen", "blue", "yellow",
    "cyan", "sienna", "orange", "coral"
};

/*
 * This widget is not self-contained, since it generates merely specifications
 * that a tcl driver is supposed to use when rendering a canvas. Therefore,
 * not all the options are actually communicated from tcl to C back to
 * tcl during the xgraph plot function. Instead, many of the attributes like
 * seriescolors are to be looked up in a configuration step prior to drawing.
 * The C database is merely a much more convenient repository than big
 * tcl variables.
 */
static
xg_configspec tkspecs_color[] = {
{TK_CONFIG_DOUBLE,"-barbase","BarBase","Jump",DEF_BAR_BASE},
{TK_CONFIG_BOOLEAN,"-bargraph","BarGraph","Boolean",DEF_BAR_FLAG},
{TK_CONFIG_PIXELS,"-barwidth","BarWidth","barWidth",DEF_BAR_WIDTH},
{TK_CONFIG_COLOR,"-bordercolor","Border","Foreground",DEF_COL_BORDER},
{TK_CONFIG_PIXELS,"-borderwidth","BorderSize","BorderWidth",DEF_BORDER_WIDTH},
{TK_CONFIG_BOOLEAN,"-boundingbox","BoundBox","Boolean",DEF_BB_FLAG},
{TK_CONFIG_PIXELS,"-gridwidth","GridSize","BorderWidth",DEF_GRID_SIZE},
{TK_CONFIG_BITMAP,"-gridstipple","GridStipple","Bitmap","tkxgraph_bits_0"},
{TK_CONFIG_PIXELS,"-height","Height","Height",DEF_HEIGHT},
{TK_CONFIG_FONT,"-labelfont","LabelFont","Font",DEF_LABEL_FONT},
{TK_CONFIG_BOOLEAN,"-logx","LogX","Boolean",DEF_LOGX_FLAG},
{TK_CONFIG_BOOLEAN,"-logy","LogY","Boolean",DEF_LOGX_FLAG},
{TK_CONFIG_BOOLEAN,"-mm","Markers","Boolean","0"},
{TK_CONFIG_BOOLEAN,"-mM","StyleMarkers","Boolean","0"},
{TK_CONFIG_BOOLEAN,"-mp","PixelMarkers","Boolean","0"},
{TK_CONFIG_BOOLEAN,"-mP","LargePixels","Boolean","0"},
{TK_CONFIG_BOOLEAN,"-negatex","NegateX","Boolean",DEF_NEGATEX_FLAG},
{TK_CONFIG_BOOLEAN,"-negatey","NegateY","Boolean",DEF_NEGATEY_FLAG},
{TK_CONFIG_BOOLEAN,"-nolines","NoLines","Boolean",DEF_NOLINE_FLAG},
{XG_CONFIG_MANY,"-serieslegend","SeriesName%d","Text","Set %d"},
{XG_CONFIG_MANY,"-seriescolor","SeriesColor%d","Foreground",""},
{XG_CONFIG_MANY,"-serieswidth","SeriesWidth%d","BorderWidth",DEF_LINE_WIDTH},
{XG_CONFIG_MANY,"-seriesmarker","SeriesMark%d","Bitmap","pixel"},
{XG_CONFIG_MANY,"-seriesstipple","SeriesStip%d","Bitmap","tkxgraph_bits_0"},
{TK_CONFIG_BOOLEAN,"-ticks","Ticks","Boolean",DEF_TICK_FLAG},
{TK_CONFIG_STRING,"-title","TitleText","Text",DEF_TITLE_TEXT},
{TK_CONFIG_FONT,"-titlefont","TitleFont","Font",DEF_TITLE_FONT},
{TK_CONFIG_PIXELS,"-width","Width","Width",DEF_WIDTH},
{TK_CONFIG_DOUBLE,"-xhigh","XHighLimit","Double",DEF_HIGH_LIMIT},
{TK_CONFIG_DOUBLE,"-xlow","XLowLimit","Double",DEF_LOW_LIMIT},
{TK_CONFIG_STRING,"-xtitle","XUnitText","Text",DEF_XUNIT_TEXT},
{TK_CONFIG_DOUBLE,"-yhigh","YHighLimit","Double",DEF_HIGH_LIMIT},
{TK_CONFIG_DOUBLE,"-ylow","YLowLimit","Double",DEF_LOW_LIMIT},
{TK_CONFIG_STRING,"-ytitle","YUnitText","Text",DEF_YUNIT_TEXT},
{TK_CONFIG_COLOR,"-zerocolor","ZeroColor","Foreground",DEF_COL_ZEROCOLOR},
{TK_CONFIG_PIXELS,"-zerowidth","ZeroWidth","BorderWidth",DEF_COL_ZEROWIDTH},
{TK_CONFIG_END,"","","",""}
};

static 
xg_configspec tkspecs_bw[] = {
{TK_CONFIG_DOUBLE,"-barbase","BarBase","Jump",DEF_BAR_BASE},
{TK_CONFIG_BOOLEAN,"-bargraph","BarGraph","Boolean",DEF_BAR_FLAG},
{TK_CONFIG_PIXELS,"-barwidth","BarWidth","barWidth",DEF_BAR_WIDTH},
{TK_CONFIG_COLOR,"-bordercolor","Border","Foreground",DEF_BW_BORDER},
{TK_CONFIG_PIXELS,"-borderwidth","BorderSize","BorderWidth",DEF_BORDER_WIDTH},
{TK_CONFIG_BOOLEAN,"-boundingbox","BoundBox","Boolean",DEF_BB_FLAG},
{TK_CONFIG_PIXELS,"-gridwidth","GridSize","BorderWidth",DEF_GRID_SIZE},
{TK_CONFIG_BITMAP,"-gridstipple","GridStipple","Bitmap","tkxgraph_bits_0"},
{TK_CONFIG_PIXELS,"-height","Height","Height",DEF_HEIGHT},
{TK_CONFIG_FONT,"-labelfont","LabelFont","Font",DEF_LABEL_FONT},
{TK_CONFIG_BOOLEAN,"-logx","LogX","Boolean",DEF_LOGX_FLAG},
{TK_CONFIG_BOOLEAN,"-logy","LogY","Boolean",DEF_LOGX_FLAG},
{TK_CONFIG_BOOLEAN,"-mm","Markers","Boolean","0"},
{TK_CONFIG_BOOLEAN,"-mM","StyleMarkers","Boolean","0"},
{TK_CONFIG_BOOLEAN,"-mp","PixelMarkers","Boolean","0"},
{TK_CONFIG_BOOLEAN,"-mP","LargePixels","Boolean","0"},
{TK_CONFIG_BOOLEAN,"-negatex","NegateX","Boolean",DEF_NEGATEX_FLAG},
{TK_CONFIG_BOOLEAN,"-negatey","NegateY","Boolean",DEF_NEGATEY_FLAG},
{TK_CONFIG_BOOLEAN,"-nolines","NoLines","Boolean",DEF_NOLINE_FLAG},
{XG_CONFIG_MANY,"-serieslegend","SeriesName%d","Text","Set %d"},
{XG_CONFIG_MANY,"-seriescolor","SeriesColor%d","Foreground",""},
{XG_CONFIG_MANY,"-serieswidth","SeriesWidth%d","BorderWidth",DEF_LINE_WIDTH},
{XG_CONFIG_MANY,"-seriesmarker","SeriesMark%d","Bitmap","pixel"},
{XG_CONFIG_MANY,"-seriesstipple","SeriesStip%d","Bitmap","tkxgraph_bits_0"},
{TK_CONFIG_BOOLEAN,"-ticks","Ticks","Boolean",DEF_TICK_FLAG},
{TK_CONFIG_STRING,"-title","TitleText","Text",DEF_TITLE_TEXT},
{TK_CONFIG_FONT,"-titlefont","TitleFont","Font",DEF_TITLE_FONT},
{TK_CONFIG_PIXELS,"-width","Width","Width",DEF_WIDTH},
{TK_CONFIG_DOUBLE,"-xhigh","XHighLimit","Double",DEF_HIGH_LIMIT},
{TK_CONFIG_DOUBLE,"-xlow","XLowLimit","Double",DEF_LOW_LIMIT},
{TK_CONFIG_STRING,"-xtitle","XUnitText","Text",DEF_XUNIT_TEXT},
{TK_CONFIG_DOUBLE,"-yhigh","YHighLimit","Double",DEF_HIGH_LIMIT},
{TK_CONFIG_DOUBLE,"-ylow","YLowLimit","Double",DEF_LOW_LIMIT},
{TK_CONFIG_STRING,"-ytitle","YUnitText","Text",DEF_YUNIT_TEXT},
{TK_CONFIG_COLOR,"-zerocolor","ZeroColor","Foreground",DEF_BW_ZEROCOLOR},
{TK_CONFIG_PIXELS,"-zerowidth","ZeroWidth","BorderWidth",DEF_BW_ZEROWIDTH},
{TK_CONFIG_END,"","","",""}
};

/* this gets assigned to be one of the above two once depth checked */
static
xg_configspec *tkspecs =NULL;



/* some forward declarations to shutup gcc*/

static int TransformCompute();
static int DrawTitle();
static int DrawLegend();
static int DrawGridAndAxis();
static int DrawData();
static void WriteValue();
static void InitSets();

static void parse_geometry();
static void rdSetName();
/* */

extern void init_X(); /* xgX.c */
extern void init_tk(); /* xgX.c */
static void do_error(); /* coded at end to fprintf */

static void set_mark_flags();
static int Traverse();

typedef struct geom_spec {
  int w;
  int h;
  int x;
  int y;
} Geom;

typedef struct point_list {
    int numPoints;		/* Number of points in group */
    int allocSize;		/* Allocated size            */
    double *xvec;		/* X values of one log state */
    double *yvec;		/* Y values of one log state */
    double *swapx;		/* X values of the other log state */
    double *swapy;		/* Y values of the other log state */
    struct point_list *next;	/* Next set of points        */
    /*
     * if wi->logxdone !=0, then xvec = log(swapx)
     * and swapx is the original data.
     * if  wi->logxdone ==0, then xvec is the original data and swapx
     * may be NULL or not depending on whether the user has ever before
     * plotted logx. ditto logy/yvec/swapy.
     */
} PointList;

typedef struct new_data_set {
    char *setName;		/* Name of data set, from pt  */
    PointList *list;		/* List of point arrays */
} NewDataSet;


/* state variables must be reset at start of read.
 */
static int setNumber = 0; /* number of set being read */
static PointList **curSpot = (PointList **) 0;
static PointList *curList = (PointList *) 0;
static int newGroup = 0;
static int redundant_set = 0;


#define HARDCOPY_IN_PROGRESS	0x01
#define TKXGMAGIC 6324851
#define TKXGBADMAGIC 8791639
typedef struct local_win {
    int magic;
    char *cmdname;		/* tk function name for this plot */
    double loX, loY, hiX, hiY;	/* Local bounding box of window         */
    int XOrgX, XOrgY;		/* Origin of bounding box on screen     */
    int XOppX, XOppY;		/* Other point defining bounding box    */
    int logxdone,logydone;      /* flags to say if logs computed for output */
    int logxtaken,logytaken;    /* flags to say if logs already computed */
    double UsrOrgX, UsrOrgY;	/* Origin of bounding box in user space */
    double UsrOppX, UsrOppY;	/* Other point of bounding box          */
    double XUnitsPerPixel;	/* X Axis scale factor                  */
    double YUnitsPerPixel;	/* Y Axis scale factor                  */
    xgOut dev_info;		/* Device information                   */
    int flags;			/* Window flags                         */
    Tcl_Interp *interp;		/* wish interpreter */
    st_table *pt;		/* parameter table */
    double llx, lly, urx, ury;  /* Bounding box of all data */
    NewDataSet PlotData[MAXSETS];	/* the data */
    XSegment *Xsegs;		/* Point space for plotters  */
    int numFiles;		/* Number of input files   */
    int nseries;		/* number of series in this data */
    char *inFileNames[MAXSETS]; 	/* File names              */
} XGDWin;
/* we are relying on this structure being calloc'd */

#define SCREENX(ws, userX) \
	(((int) (((userX) - ws->UsrOrgX)/ws->XUnitsPerPixel + 0.5)) + ws->XOrgX)
#define SCREENY(ws, userY) \
	(ws->XOppY - ((int) (((userY) - ws->UsrOrgY)/ws->YUnitsPerPixel + 0.5)))

/* Other globally set defaults */

int tkxg_depth;			/* Depth of screen         */



static int Traverse(flags,wi)
int flags;			/* Options */
XGDWin *wi;
/*
 * Traverses through all of the data applying certain options to the
 * data and computing the overall bounding box.  The flags are:
 *   LOG_X	Take the log of the X axis
 *   LOG_Y	Take the log of the Y axis
 * This function is in charge of swapping x,y and logx,logy. If
 * The log cannot be taken, this will leave the data in the unlogged state.
 * logxdone and logydone will reveal what has happened.
 * logx,ytaken reveal whether a computation is needed.
 * Returns tcl_ok if all is well.
 */
{
    int i, j;
    unsigned int checky, checkx, erry=0,errx=0,converty,convertx,yswap,xswap;
    PointList *spot;
    double *hold;

   /* see if a transformation is needed and not already done */

    checky = ((flags & LOG_Y) && !(wi->logytaken));
    checkx = ((flags & LOG_X) && !(wi->logxtaken));
    if (checkx || checky) {
        for (i = 0;  i < MAXSETS;  i++) {
        /* forall sets */
	    for (spot = wi->PlotData[i].list;  spot;  spot = spot->next) {
            /* forall lists in set */
	        for (j = 0;  j < spot->numPoints;  j++) {
                /* forall points in list */
		    if (checky && spot->yvec[j] <= 0.0) {
                        erry++;
		    }
		    if (checkx && spot->xvec[j] <= 0.0) {
                        errx++;
		    }
                    if (errx && erry) {
                        break;
                    }
	        }
                if (errx && erry) {
                    break;
                }
	    }
            if (errx && erry) {
                break;
            }
        }
    }

    /* see if conversion is possible and needed or unconversion needed.
     * then set the flags to indicate what has been (will be) done.
     */
    /* the NULL check is possibly redundant and certainly heuristic */
    converty = (checky && !erry);
    convertx = (checkx && !errx);
    /* see if a swap is needed. 
     */
    yswap = 0;
    if (converty) {
        /* first time requested and approved */
        yswap = 1;
    } else {
        if ( (wi->logydone && !(flags & LOG_Y) ) ||
             (!(wi->logydone) && (flags & LOG_Y) && (wi->logytaken)) ) {
            yswap = 1;
        }
    }
    xswap = 0;
    if (convertx) {
        /* first time requested and approved */
        xswap = 1;
    } else {
        if ( (wi->logxdone && !(flags & LOG_X) ) ||
             (!(wi->logxdone) && (flags & LOG_X) && (wi->logxtaken)) ) {
            xswap = 1;
        }
    }


    if (converty) {
        wi->logytaken = 1;
    }
    if (convertx) {
        wi->logxtaken = 1;
    }
    if (!(flags & LOG_Y)) {
        wi->logydone = 0;
    }
    if (!(flags & LOG_X)) {
        wi->logxdone = 0;
    }
    if ((flags & LOG_Y) && wi->logytaken) {
        wi->logydone = 1;
    }
    if ((flags & LOG_X) && wi->logxtaken) {
        wi->logxdone = 1;
    }
    for (i = 0;  i < MAXSETS;  i++) {
    /* forall sets */
	for (spot = wi->PlotData[i].list;  spot;  spot = spot->next) {
        /* forall lists in set */
            if (converty) {
                assert(yswap);
                spot->swapy =
                    (double *)malloc(sizeof(double)*(spot->numPoints));
	        for (j = 0;  j < spot->numPoints;  j++) {
                /* forall points in list */
		    spot->swapy[j] = log10(spot->yvec[j]);
                }
            }
            if (convertx) {
                assert(xswap);
                spot->swapx =
                    (double *)malloc(sizeof(double)*(spot->numPoints));
	        for (j = 0;  j < spot->numPoints;  j++) {
                /* forall points in list */
		    spot->swapx[j] = log10(spot->xvec[j]);
                }
            }
            if (yswap) {
                hold = spot->swapy;
                spot->swapy = spot->yvec; 
                spot->yvec = hold;
            }
            if (xswap) {
                hold = spot->swapx;
                spot->swapx = spot->xvec; 
                spot->xvec = hold;
            }
        }
    }
    /* log or not, xvec, yvec now have plotted data */
    /* derive bounding box of all data */
    for (i = 0;  i < MAXSETS;  i++) {
    /* forall sets */
	for (spot = wi->PlotData[i].list;  spot;  spot = spot->next) {
        /* forall lists in set */
	    for (j = 0;  j < spot->numPoints;  j++) {
            /* forall points in list */
		/* Update global bounding box */
		if (spot->xvec[j] < wi->llx) wi->llx = spot->xvec[j];
		if (spot->xvec[j] > wi->urx) wi->urx = spot->xvec[j];
		if (spot->yvec[j] < wi->lly) wi->lly = spot->yvec[j];
		if (spot->yvec[j] > wi->ury) wi->ury = spot->yvec[j];
	    }
	}
    }
    return TCL_OK;
}





static
void Tkxg_PlotSize(wi,lowX, lowY, upX, upY, asp,interp)
XGDWin *wi;
double lowX, lowY;		/* Lower left corner  */
double upX, upY;		/* Upper right corner */
double asp;			/* Aspect ratio       */
Tcl_Interp *interp;
/* 
 * insure sanity after user input of bounds, calculate
 * padding, width, height. Do this before drawing.
 * call this and tk_reset between parseargs of any form
 * and drawwindow.
 */ 
{    
    double pad;
    int width, height;
	/* if user bounds are screwy, use data bounds instead */
    (void)interp;
    if (upX > lowX) {
	wi->loX = lowX;
	wi->hiX = upX;
    } else {
	wi->loX = wi->llx;
	wi->hiX = wi->urx;
    }
    if (upY > lowY) {
	wi->loY = lowY;
	wi->hiY = upY;
    } else {
	wi->loY = wi->lly;
	wi->hiY = wi->ury;
    }

    /* Increase the padding for aesthetics, disallow 0x0 box */
    if (wi->hiX - wi->loX == 0.0) {
	pad = MAX(0.5, fabs(wi->hiX/2.0));
	wi->hiX += pad;
	wi->loX -= pad;
    }
    if (wi->hiY - wi->loY == 0) {
	pad = MAX(0.5, fabs(wi->ury/2.0));
	wi->hiY += pad;
	wi->loY -= pad;
    }

    /* Add 10% padding to bounding box (div by 20 yeilds 5%) */
    pad = (wi->hiX - wi->loX) / 20.0;
    wi->loX -= pad;  wi->hiX += pad;
    pad = (wi->hiY - wi->loY) / 20.0;
    wi->loY -= pad;  wi->hiY += pad;

    /* Aspect ratio computation */
    if (asp < 1.0) {
	height = NORMSIZE;
	width = ((int) (((double) NORMSIZE) * asp));
    } else {
	width = NORMSIZE;
	height = ((int) (((double) NORMSIZE) / asp));
    }
    height = MAX(MINDIM, height);
    width = MAX(MINDIM, width);
}



static
void InitSets(pt,interp,wi)
st_table **pt;
Tcl_Interp *interp;
XGDWin *wi;
/*
 * Initializes the data sets with default information.  Sets up
 * original values for parameters in parameters package.
 * This is a per graph file function.
 */
{
    int idx;
    char buf[1024];

    /* set up the hash table for all the option data */
    *pt = param_init(interp);

    param_set(*pt,"Debug", BOOL, "false");
    param_set(*pt,"Geometry", STR, DEF_GEOMETRY);
    param_set(*pt,"ReverseVideo", BOOL, DEF_REVERSE);

    param_set(*pt,"Width", INT, DEF_WIDTH);
    param_set(*pt,"Height", INT, DEF_HEIGHT);
    param_set(*pt,"BorderSize", INT, DEF_BORDER_WIDTH);
    param_set(*pt,"TitleText", STR, DEF_TITLE_TEXT);
    param_set(*pt,"XUnitText", STR, DEF_XUNIT_TEXT);
    param_set(*pt,"YUnitText", STR, DEF_YUNIT_TEXT); /* YUnits */
    param_set(*pt,"Ticks", BOOL, DEF_TICK_FLAG);

    param_set(*pt,"Markers", BOOL, DEF_MARKER_FLAG); /* markFlag (-m) */
    param_set(*pt,"StyleMarkers", BOOL, DEF_DIFFMARK_FLAG); /* colorMark(-M)*/
    param_set(*pt,"PixelMarkers", BOOL, DEF_PIXMARK_FLAG); /* pixelMarks(-p)*/
    param_set(*pt,"LargePixels", BOOL, DEF_LARGEPIX_FLAG); /* bigPixel (-P)*/

    param_set(*pt,"BoundBox", BOOL, DEF_BB_FLAG);
    param_set(*pt,"NoLines", BOOL, DEF_NOLINE_FLAG);
    param_set(*pt,"LogX", BOOL, DEF_LOGX_FLAG);
    param_set(*pt,"LogY", BOOL, DEF_LOGY_FLAG); /* logYFlag */
    param_set(*pt,"NegateX", BOOL, DEF_NEGATEX_FLAG);
    param_set(*pt,"NegateY", BOOL, DEF_NEGATEY_FLAG);
    param_set(*pt,"BarGraph", BOOL, DEF_BAR_FLAG);
    param_set(*pt,"BarBase", DBL, DEF_BAR_BASE);
    param_set(*pt,"BarWidth", DBL, DEF_BAR_WIDTH);
    param_set(*pt,"LineWidth", INT, DEF_LINE_WIDTH);
    param_set(*pt,"GridSize", INT, DEF_GRID_SIZE);
    param_set(*pt,"GridStyle", STYLE, DEF_GRID_STYLE);
    {
        param_style ps;
        ps = PM_STYLE(*pt,"GridStyle",interp);
        param_set(*pt,"GridStipple", STR, ps.tkbitname);
    }

    param_set(*pt,"Device", STR, DEF_DEVICE);
    param_set(*pt,"Disposition", STR, DEF_DISPOSITION);
    param_set(*pt,"FileOrDev", STR, DEF_FILEORDEV);

    /* Set the user bounding box */
    param_set(*pt,"XLowLimit", DBL, DEF_LOW_LIMIT);
    param_set(*pt,"YLowLimit", DBL, DEF_LOW_LIMIT);
    param_set(*pt,"XHighLimit", DBL, DEF_HIGH_LIMIT);
    param_set(*pt,"YHighLimit", DBL, DEF_HIGH_LIMIT);


    /* Depends critically on whether the display has color */
    if (tkxg_depth < 4) {
	/* Its black and white */
	param_set(*pt,"Background", PIXEL, DEF_BW_BACKGROUND);
	param_set(*pt,"Border", PIXEL, DEF_BW_BORDER);
	param_set(*pt,"ZeroColor", PIXEL, DEF_BW_ZEROCOLOR);
	param_set(*pt,"ZeroWidth", INT, DEF_BW_ZEROWIDTH);
	param_set(*pt,"ZeroStyle", STYLE, DEF_BW_ZEROSTYLE);
	param_set(*pt,"Foreground", PIXEL, DEF_BW_FOREGROUND);
	/* Initialize set defaults */
	for (idx = 0;  idx < MAXATTR;  idx++) {
	    (void) sprintf(buf, "%d.Style", idx);
	    param_set(*pt,buf, STYLE, defStyle[idx]);
	    (void) sprintf(buf, "%d.Color", idx);
	    param_set(*pt,buf, PIXEL, DEF_BW_FOREGROUND);
	}
    } else {
	/* Its color */
	param_set(*pt,"Background", PIXEL, DEF_COL_BACKGROUND);
	param_set(*pt,"Border", PIXEL, DEF_COL_BORDER);
	param_set(*pt,"ZeroColor", PIXEL, DEF_COL_ZEROCOLOR);
	param_set(*pt,"ZeroWidth", INT, DEF_COL_ZEROWIDTH);
	param_set(*pt,"ZeroStyle", STYLE, DEF_COL_ZEROSTYLE);
	param_set(*pt,"Foreground", PIXEL, DEF_COL_FOREGROUND);
	/* Initalize attribute colors defaults */
	for (idx = 0;  idx < MAXATTR;  idx++) {
	    (void) sprintf(buf, "%d.Style", idx);
	    param_set(*pt,buf, STYLE, defStyle[idx]);
	    (void) sprintf(buf, "%d.Color", idx);
	    param_set(*pt,buf, PIXEL, defColors[idx]);
	}
    }

    param_set(*pt,"LabelFont", FONT, DEF_LABEL_FONT);
    param_set(*pt,"TitleFont", FONT, DEF_TITLE_FONT);

    {
        char seriesString[80];
        for (idx=0;idx <MAXSETS; idx++) {
            sprintf(seriesString,"SeriesWidth%d",idx);
            param_set(*pt,seriesString,INT,DEF_LINE_WIDTH);

            sprintf(seriesString,"SeriesMark%d",idx);
            param_set(*pt,seriesString,STR,"pixel");

            sprintf(seriesString,"SeriesColor%d",idx);
	    param_set(*pt,seriesString,STR, 
                (tkxg_depth >3) ? defColors[idx % MAXATTR] :
                                  DEF_BW_FOREGROUND);
 
            sprintf(seriesString,"SeriesStip%d",idx);
            sprintf(buf,"%s%d",TKXGBITMAPROOT,idx % MAXATTR);
            param_set(*pt,seriesString,STR,buf);
        }
    }
    /* Initialize the data sets to empty with dumb names */
    for (idx = 0;  idx < MAXSETS;  idx++) {
	(void) sprintf(buf, "Set %d", idx);
        setNumber = idx; /* set global */
        rdSetName(buf,wi,idx);
	wi->PlotData[idx].list = (PointList *) 0;
    }
    setNumber = 0; /* reset global */
}


XGDWin *NewWindow(interp)
Tcl_Interp *interp;	/* wish interpreter */
/*
 * Creates and a new graph space.  This includes allocating its
 * local structures and associating it with the interp.
 */
{
    XGDWin *new;
    new = (XGDWin *) calloc(1,sizeof(XGDWin));
    new->magic = TKXGMAGIC;
    /* Set up hard-wired defaults and allocate spaces */
    InitSets(&(new->pt),interp,new);

    new->interp = interp;

    /* define basic tk out_info device. There are other
     * possible devices (idraw, eps, hpgl)
     */
    set_tk(&(new->dev_info),interp,new->pt);
    return new;
}


#if 0 /* need to connect somewhere */
static
int arghelp(i)
Tcl_Interp *i;
{
#define FS(str)	Tcl_AppendResult(i, str,(char *)NULL)
    FS("format: tkxg [-<digit> set_name] [-bar] [-bb] [-bd border_color]\n");
    FS("       [-bg background_color] [-brb bar_base] [-brw bar_width]\n");
    FS("       [-bw bdr_width] [-db] [-fg foreground_color] [-gw grid_size]\n");
    FS("       [-gs grid_style] [-lf label_font] [-lnx] [-lny]\n");
    FS("       [-lw line_width] [-lx x1,x2] [-ly y1,y2] [-m] [-M]\n");
    FS("       [-nl] [-nx] [-ny] [-p] [-P] [-rv]\n");
    FS("       [-t title] [-tf title_font] [-tk] [-x x_unit_name]\n");
    FS("       [-y y_unit_name] [-zg zero_color] [-zw zero_size]\n");
    FS("       [=WxH+X+Y] [-display <host>:<disp>.<screen>] file...\n\n");
    FS("-bar   Draw bar graph with base -brb and width -brw\n");
    FS("-bb    Draw bounding box around data\n");
    FS("-db    Turn on debugging\n");
    FS("-lnx   Logarithmic scale for X axis\n");
    FS("-lny   Logarithmic scale for Y axis\n");
    FS("-m -M  Mark points distinctively (M varies with color)\n");
    FS("-nl    Don't draw lines (scatter plot)\n");
    FS("-nx    Flip x axis signs\n");
    FS("-ny    Flip y axis signs\n");
    FS("-p -P  Mark points with dot (P means big dot)\n");
    FS("-rv    Reverse video on black and white displays(implemented?)\n");
    FS("-tk    Draw tick marks instead of full grid\n");
    return TCL_ERROR;
#undef FS
}
#endif

int argerror(err, val,i)
char *err, *val;
Tcl_Interp *i;
{
    Tcl_AppendResult(i,"Error: ",val,": ",err,"\n\n",(char *)NULL);
    return TCL_ERROR;
}


#define MAXLO	30

int ParseArgs(argc, argv, do_it,interp,pt,wi)
int argc;
char *argv[];
int do_it;
Tcl_Interp *interp;
st_table *pt;
XGDWin *wi;
/*
 * This routine parses the argument list for xgraph.  There are too
 * many to mention here so I won't. 
 * If `do_it' is 1, options are actually changed and files ignored.  
 * If `do_it' is 0, the argument list
 * is parsed but the options aren't set, instead only
 * file names are collected. 
 * If do_it is 2, all arguments are expected to be filenames without
 * any switches to parse and files are added to the list.
 * Returns tclok or tclerror depending on results.
 * This assumes that all of argv is arguments and that argv0 is not
 * special.
 */
{
#define ARG(opt, name) \
  if (strcmp(argv[idx], opt) == 0) { \
      if (do_it) param_set(pt,name, BOOL, "on"); \
      idx++; continue; \
  }

#define ARG2(opt, name, type, missing) \
  if (strcmp(argv[idx], opt) == 0) { \
     if (idx+1 >= argc) return argerror(missing, argv[idx],interp); \
     if (do_it) param_set(pt,name, type, argv[idx+1]); \
     idx += 2; continue;\
  }

/* sets item %name%(number in idx+1) to idx+2. name < 35 characters.
 * handles options which come in cliques differing only by number.
 */
#define ARG3(opt, name, type, missing) \
  if (strcmp(argv[idx], opt) == 0) { \
     if (idx+2 >= argc) return argerror(missing, argv[idx],interp); \
     if (do_it) { \
         char seriesString[60]; \
         param_set(pt,"SeriesNumber",INT, argv[idx+1]); \
         sprintf(seriesString,"%s%d",name, \
                 PM_INT(wi->pt,"SeriesNumber",interp)); \
         param_set(pt,seriesString,type,argv[idx+2]); \
     } \
     idx += 3; continue; \
  }

    int idx, set;
    int stdin_count = 1; /* no stdin input allowed */
    char *hi;

    if (do_it == 2) {
        idx = 0;
        while (idx < argc && idx < MAXSETS && wi->numFiles < MAXSETS) {
            wi->inFileNames[wi->numFiles] = STRDUP(argv[idx]);
            wi->numFiles++;
            idx++;
        }
        return TCL_OK;
    }

    idx = 0;
    while (idx < argc) {
	if (argv[idx][0] == '-' && argv[idx][1] != '\0') {
	    /* Check to see if its a data set name */
	    if (sscanf(argv[idx], "-%d", &set) == 1) {
		/* The next string is a set name */
		if (idx+1 >= argc)  {
                    return argerror("missing set name", argv[idx],interp);
                }
		if (do_it) {
                    char sername[60];
                    sprintf(sername,"%s%d","SeriesName",set);
                    param_set(pt,sername,STR,argv[idx+1]); 
		    wi->PlotData[set].setName = PM_STR(wi->pt,sername,interp);
		}
		idx += 2;
	    } else {
               /* data file option  -f -file */
                if (strcmp(argv[idx], "-f") == 0 || 
                    strcmp(argv[idx], "-file") == 0) {
                     if (idx+1 >= argc)  {
                          return argerror("missing file name",
                                           argv[idx],interp);
                     }
                     if (!do_it) {
	                 wi->inFileNames[wi->numFiles] = STRDUP(argv[idx+1]);
	                 wi->numFiles++;
                     }
                     idx += 2;
                     continue;
                }

		/* Some non-dataset option */
		ARG2("-x", "XUnitText", STR, "missing axis name");
		ARG2("-xtitle", "XUnitText", STR, "missing axis name");
		ARG2("-y", "YUnitText", STR, "missing axis name");
		ARG2("-ytitle", "YUnitText", STR, "missing axis name");
		ARG2("-t", "TitleText", STR, "missing plot title");
		ARG2("-title", "TitleText", STR, "missing plot title");
		ARG2("-fg", "Foreground", PIXEL, "missing color name");
		ARG2("-foreground", "Foreground", PIXEL, "missing color name");
		ARG2("-bg", "Background", PIXEL, "missing color name");
		ARG2("-background", "Background", PIXEL, "missing color name");
		ARG2("-bd", "Border", PIXEL, "missing color name");
		ARG2("-bordercolor", "Border", PIXEL, "missing color name");
		ARG2("-bw", "BorderSize", INT, "missing border size");
		ARG2("-borderwidth", "BorderSize", INT, "missing border size");
		ARG2("-zg", "ZeroColor", PIXEL, "missing color name");
		ARG2("-zerocolor", "ZeroColor", PIXEL, "missing color name");
		ARG2("-zw", "ZeroWidth", INT, "missing width");
		ARG2("-zerowidth", "ZeroWidth", INT, "missing width");
		ARG2("-tf", "TitleFont", FONT, "missing font name");
		ARG2("-titlefont", "TitleFont", FONT, "missing font name");
		ARG2("-lf", "LabelFont", FONT, "missing font name");
		ARG2("-labelfont", "LabelFont", FONT, "missing font name");
		ARG("-rv", "ReverseVideo");
		ARG("-tk", "Ticks");
		ARG2("-ticks", "Ticks",BOOL,"missing boolean value");
		ARG("-bb", "BoundBox");
		ARG2("-boundingbox", "BoundBox", BOOL,"missing boolean value");
		ARG2("-xlow", "XLowLimit", DBL, "missing float value");
		ARG2("-xhigh", "XHighLimit", DBL, "missing float value");
		if (strcmp(argv[idx], "-lx") == 0) {
		    /* Limit the X coordinates */
		    if (idx+1 >= argc) {
                        return argerror("missing coordinate(s)",
					argv[idx],interp);
                    }
		    if ( (hi = strchr(argv[idx+1], ','))!=NULL) {
			char low[MAXLO];
                        unsigned delta;
		        delta =  hi-argv[idx+1];
			(void) strncpy(low, argv[idx+1], delta);
			low[delta] = '\0';
			hi++;
			if (do_it) {
			    param_set(pt,"XLowLimit", DBL, argv[idx+1]);
			    param_set(pt,"XHighLimit", DBL, hi);
			}
		    } else {
			return argerror("limit coordinates not 'lo,hi'",
				 argv[idx],interp);
		    }
		    idx += 2;
		    continue;
		}
		ARG2("-ylow", "YLowLimit", DBL, "missing coordinate value");
		ARG2("-yhigh", "YHighLimit", DBL, "missing coordinate value");
		if (strcmp(argv[idx], "-ly") == 0) {
		    /* Limit the Y coordinates */
		    if (idx+1 >= argc) {
                        return argerror("missing coordinate(s)",
					 argv[idx],interp);
                    }
		    if ((hi = strchr(argv[idx+1], ','))!=NULL) {
			char low[MAXLO];
			unsigned delta;
			delta = hi-argv[idx+1];
			(void) strncpy(low, argv[idx+1], delta);
			low[delta] = '\0';
			hi++;
			if (do_it) {
			    param_set(pt,"YLowLimit", DBL, argv[idx+1]);
			    param_set(pt,"YHighLimit", DBL, hi);
			}
		    } else {
			return argerror("limit coordinates not 'lo,hi'",
				 argv[idx],interp);
		    }
		    idx += 2;
		    continue;
		} 
		ARG2("-lw", "LineWidth", INT, "missing line width");
                /* see also serieswidth below */
		ARG2("-linewidth","LineWidth", INT, "missing line width");
		ARG("-nl", "NoLines");
		ARG2("-nolines", "NoLines",BOOL,"missing boolean value");
		ARG("-m", "Markers");
		ARG("-M", "StyleMarkers");
		ARG("-p", "PixelMarkers");
		ARG("-P", "LargePixels");
                /* toggles for same */
		ARG2("-mm", "Markers",BOOL, "missing boolean value");
		ARG2("-mM", "StyleMarkers",BOOL, "missing boolean value");
		ARG2("-mp", "PixelMarkers",BOOL, "missing boolean value");
		ARG2("-mP", "LargePixels",BOOL, "missing boolean value");
		ARG("-lnx", "LogX");
		ARG2("-logx", "LogX", BOOL, "missing boolean value");
		ARG("-lny", "LogY");
		ARG2("-logy", "LogY", BOOL, "missing boolean value");
		ARG("-ny", "NegateY");
		ARG2("-negatey", "NegateY", BOOL, "missing boolean value");
		ARG("-nx", "NegateX");
		ARG2("-negatex", "NegateX", BOOL, "missing boolean value");
		ARG("-bar", "BarGraph");
		ARG2("-bargraph", "BarGraph", BOOL, "missing boolean value");
		ARG2("-brw", "BarWidth", DBL, "missing width");
		ARG2("-barwidth", "BarWidth", DBL, "missing width");
		ARG2("-brb", "BarBase", DBL, "missing base");
		ARG2("-barbase", "BarBase", DBL, "missing base");
		ARG("-db", "Debug");
		ARG2("-gw", "GridSize", INT, "missing grid width");
		ARG2("-gridwidth", "GridSize", INT, "missing grid width");
		ARG2("-gs", "GridStyle",STYLE,"missing grid style");
		ARG2("-gridstyle", "GridStyle",STYLE,"missing grid style");
		ARG2("-geometry", "Geometry",STR,"missing geometry");
                /* can't arg2 width because it must update geometry also */
                if (strcmp(argv[idx], "-width") == 0) { 
                    if (idx+1 >= argc) {
                        return argerror("missing width",argv[idx],interp);
                    }
                    if (do_it) {
                       Geom geom;
                       char newgeom[160];
                       parse_geometry(&geom,PM_STR(wi->pt,"Geometry",interp));
                       param_set(pt,"Width", INT, argv[idx+1]);
                       geom.w = PM_INT(wi->pt,"Width",interp);
                       sprintf(newgeom,"%dx%d+%d+%d",geom.w,geom.h,
                               geom.x,geom.y);
                       param_set(pt,"Geometry",STR,newgeom);
                    }
                    idx += 2; 
                    continue;
                }
                /* can't arg2 height because it must update geometry also */
                if (strcmp(argv[idx], "-height") == 0) { 
                    if (idx+1 >= argc) {
                        return argerror("missing height",argv[idx],interp);
                    }
                    if (do_it) {
                       Geom geom;
                       char newgeom[160];
                       parse_geometry(&geom,PM_STR(wi->pt,"Geometry",interp));
                       param_set(pt,"Height", INT, argv[idx+1]);
                       geom.h = PM_INT(wi->pt,"Height",interp);
                       sprintf(newgeom,"%dx%d+%d+%d",geom.w,geom.h,
                               geom.x,geom.y);
                       param_set(pt,"Geometry",STR,newgeom);
                    }
                    idx += 2; 
                    continue;
                }
                /* ignore display specifications */
		if (strcmp(argv[idx], "-display") == 0) {
		    idx += 2;
		    continue;
		}
		ARG2("-barbase","BarBase", DBL, "missing base");
                if (strcmp(argv[idx], "-serieslegend") == 0) {
                    if (idx+2 >= argc) {
                        return argerror("missing legend", argv[idx],interp); 
                    }
                    if (do_it) { 
                        char seriesString[60];
                        param_set(pt,"SeriesNumber",INT, argv[idx+1]); 
                        sprintf(seriesString,"%s%d","SeriesName",
                                   PM_INT(wi->pt,"SeriesNumber",interp)); 
                        param_set(pt,seriesString,STR,argv[idx+2]);
                        wi->PlotData[PM_INT(wi->pt,"SeriesNumber",interp)].
                            setName = PM_STR(wi->pt,seriesString,interp); 
                    } 
                    idx += 3; continue;
                }
                ARG3("-seriescolor","SeriesColor",STR,"missing colorname");
                ARG3("-serieswidth","SeriesWidth",INT,"missing line width");
                ARG3("-seriesmarker","SeriesMark",STR,"missing mark name");
                ARG3("-seriesstipple","SeriesStip",STR,"missing stipple");
		return argerror("unknown option", argv[idx],interp);
	    }
	} else if (argv[idx][0] == '=') {
	    /* Its an old geometry specification */
	    if (do_it)  {
                param_set(pt,"Geometry", STR, argv[idx]+1);
            }
	    idx++;
	} else {
            return argerror("unknown option",argv[idx],interp);
	}
    }
    return TCL_OK;
#undef ARG
#undef ARG2
#undef ARG3
}


/*
 * New dataset reading code
 */

static int rdSet(fn,wi)
char *fn;			/* Reading from file `fn' */
XGDWin *wi;
/*
 * Set up new dataset.  Will return zero if there are too many data sets.
 * if setname is repeated, ignores second occurence.
 */
{
    char setname[100];

    if (!redundant_set) {
	if (setNumber < MAXSETS) {
	    (void) sprintf(setname, "Set %d", setNumber);
	    if ((strcmp(wi->PlotData[setNumber].setName, setname) == 0) && 
                fn) {
		rdSetName(fn,wi,setNumber); /*baa?*/
	    }
	    curSpot = &(wi->PlotData[setNumber].list);
	    wi->PlotData[setNumber].list = (PointList *) 0;
	    newGroup = 1;
	    setNumber++;
	    redundant_set = 1;
	    return 1;
	} else {
	    return 0;
	}
    } else {
	return 1;
    }
}

/* uses a global set index */
static void rdSetName(name,wi,n)
char *name;			/* New set name */
XGDWin *wi;
int n;
/*
 * Sets the name of a data set.  Automatically makes a copy.
 */
{
    char seriesName[40];
    sprintf(seriesName,"SeriesName%d",n);
    param_set(wi->pt,seriesName,STR,name);
    wi->PlotData[n].setName = 
        PM_STR(wi->pt,seriesName,wi->interp);
}

static void rdGroup()
/*
 * Set up for reading new group of points within a dataset.
 */
{
    newGroup = 1;
}

static void rdPoint(xval, yval)
double xval, yval;		/* New point         */
/*
 * Adds a new point to the current group of the current
 * data set.
 */
{
    if (newGroup) {
	*curSpot = (PointList *) malloc(sizeof(PointList));
	curList = *curSpot;
	curSpot = &(curList->next);
	curList->numPoints = 0;
	curList->allocSize = INITSIZE;
	curList->xvec = (double *) 
              malloc((unsigned)(INITSIZE * sizeof(double)));
	curList->yvec = (double *)
              malloc((unsigned)(INITSIZE * sizeof(double)));
	curList->next = (PointList *) 0;
	newGroup = 0;
    }
    if (curList->numPoints >= curList->allocSize) {
	curList->allocSize *= 2;
	curList->xvec = (double *) realloc((char *) curList->xvec,
					   (unsigned) (curList->allocSize *
						       sizeof(double)));
	curList->yvec = (double *) realloc((char *) curList->yvec,
					   (unsigned) (curList->allocSize *
						       sizeof(double)));
    }

    curList->xvec[curList->numPoints] = xval;
    curList->yvec[curList->numPoints] = yval;

    (curList->numPoints)++;
    redundant_set = 0;
}

static int rdFindMax(wi)
XGDWin *wi;
/* 
 * Returns the maximum number of items in any one group of any
 * data set.
 */
{
    int i;
    PointList *list;
    int max = -1;

    for (i = 0;  i < setNumber;  i++) {
	for (list = wi->PlotData[i].list;  list;  list = list->next) {
	    if (list->numPoints > max) max = list->numPoints;
	}
    }
    return max;
}


typedef enum line_type {
    EMPTY, COMMENT, SETNAME, DRAWPNT, MOVEPNT, SETPARAM, ERROR
} LineType;

typedef struct point_defn {
    double xval, yval;
} Point;

typedef struct parmval_defn {
    char *name, *value;
} ParmVals;

typedef struct line_info {
    LineType type;
    union val_defn {
	char *str;		/* SETNAME, ERROR   */
	Point pnt;		/* DRAWPNT, MOVEPNT */
	ParmVals parm;		/* SETPARAM         */
    } val;
} LineInfo;

static int point_count = 0;

static LineType parse_line(line, result)
char *line;			/* Line to parse   */
LineInfo *result;		/* Returned result */
/*
 * Parses `line' into one of the types given in the definition
 * of LineInfo.  The appropriate values are filled into `result'.
 * Below are the current formats for each type:
 *   EMPTY:	All white space
 *   COMMENT:	Starts with "#"
 *   SETNAME:	A name enclosed in double quotes
 *   DRAWPNT:	Two numbers optionally preceded by keyword "draw"
 *   MOVEPNT:	Two numbers preceded by keyword "move"
 *   SETPARAM:  Two non-null strings separated by ":"
 *   ERROR:	Not any of the above (an error message is returned)
 * Note that often the values are pointers into the line itself
 * and should be copied if they are to be used over a long period.
 */
{
    char *first;

    /* Find first non-space character */
    while (*line && isspace(*line)) line++;
    if (*line) {
	if (*line == '#') {
	    /* comment */
	    result->type = COMMENT;
	} else if (*line == '"') {
	    /* setname */
	    result->type = SETNAME;
	    line++;
	    result->val.str = line;
	    while (*line && (*line != '\n') && (*line != '"')) line++;
	    if (*line) *line = '\0';
	} else {
	    first = line;
	    while (*line && !isspace(*line)) line++;
	    if (*line) {
		*line = '\0';
		if (stricmp(first, "move") == 0) {
		    /* MOVEPNT */
		    if (sscanf(line+1, "%lf %lf",
			       &result->val.pnt.xval,
			       &result->val.pnt.yval) == 2) {
			result->type = MOVEPNT;
		    } else if (sscanf(line+1, "%lf",
				      &result->val.pnt.yval) == 1) {
			result->val.pnt.xval = (double) point_count++;
			result->type = MOVEPNT;
		    } else {
			result->type = ERROR;
			result->val.str = "Cannot read move coordinates";
		    }
		} else if (stricmp(first, "draw") == 0) {
		    /* DRAWPNT */
		    if (sscanf(line+1, "%lf %lf",
			       &result->val.pnt.xval,
			       &result->val.pnt.yval) == 2) {
			result->type = DRAWPNT;
		    } else if (sscanf(line+1, "%lf",
				      &result->val.pnt.yval) == 1) {
			result->val.pnt.xval = (double) point_count++;
			result->type = DRAWPNT;
		    } else {
			result->type = ERROR;
			result->val.str = "Cannot read draw coordinates";
		    }
		} else if (first[strlen(first)-1] == ':') {
		    /* SETPARAM */
		    first[strlen(first)-1] = '\0';
		    result->val.parm.name = first;
		    line++;
		    while (*line && isspace(*line)) line++;
		    /* may be a \n at end of it */
		    if (line[strlen(line)-1] == '\n') {
			line[strlen(line)-1] = '\0';
		    }
		    result->val.parm.value = line;
		    result->type = SETPARAM;
		} else if (sscanf(first, "%lf", &result->val.pnt.xval) == 1) {
		    /* DRAWPNT */
		    if (sscanf(line+1, "%lf", &result->val.pnt.yval) == 1) {
			result->type = DRAWPNT;
		    } else {
			result->val.pnt.yval = result->val.pnt.xval;
			result->val.pnt.xval = (double) point_count++;
			result->type = DRAWPNT;
		    }
		} else {
		    /* ERROR */
		    result->type = ERROR;
		    result->val.str = "Unknown line type";
		}
	    } else {
		/* ERROR */
		result->type = ERROR;
		result->val.str = "Premature end of line";
	    }
	}
    } else {
	/* empty */
	result->type = EMPTY;
    }
    return result->type;
}


int ReadData(stream, filename,pt,wi)
FILE *stream;
char *filename;
st_table *pt;
XGDWin *wi;
/*
 * Reads in the data sets from the supplied stream.  If the format
 * is correct,  it returns the current maximum number of points across
 * all data sets.  If there is an error,  it returns -1.
 */
{
    char buffer[MAXBUFSIZE];
    LineInfo info;
    int line_count = 0;
    int errors = 0;
    
    assert(filename!=NULL && pt !=NULL && wi != NULL && stream!=NULL);
    if (!rdSet(filename,wi)) {
	(void) fprintf(stderr, "Error in file `%s' at line %d:\n  %s\n",
		       filename, line_count,
		       "Too many data sets - extra data ignored");
	return -1;
    }
    point_count = 0;
    while (fgets(buffer, MAXBUFSIZE, stream)) {
	line_count++;
	switch (parse_line(buffer, &info)) {
	case EMPTY:
	    if (!rdSet(filename,wi)) {
		(void) fprintf(stderr, "Error in file `%s' at line %d:\n  %s\n",
			       filename, line_count,
			       "Too many data sets - extra data ignored");
		return -1;
	    }
	    break;
	case COMMENT:
	    /* nothing */
	    break;
	case SETNAME:
	    rdSetName(info.val.str,wi,setNumber-1); /* baa - num files read? */
	    break;
	case DRAWPNT:
	    rdPoint(info.val.pnt.xval, info.val.pnt.yval);
	    break;
	case MOVEPNT:
	    rdGroup();
	    rdPoint(info.val.pnt.xval, info.val.pnt.yval);
	    break;
	case SETPARAM:
	    param_reset(pt,info.val.parm.name, info.val.parm.value);
	    break;
	default:
	    if (filename) {
		(void) fprintf(stderr, "Error in file `%s' at line %d:\n  %s\n",
			       filename, line_count, info.val.str);
		errors++;
	    }
	    break;
	}
    }
    if (errors) return -1; else return rdFindMax(wi);
}




int DrawWindow(win_info)
XGDWin *win_info;		/* Window information */
/*
 * Draws the data in the window.  Does not clear the window.
 * The data is scaled so that all of the data will fit.
 * Grid lines are drawn at the nearest power of 10 in engineering
 * notation.  Draws axis numbers along bottom and left hand edges.
 * Centers title at top of window.
 */
{
    int err=0;
    /* Figure out the transformation constants */
    if (TransformCompute(win_info)) {
	
	/* Draw the title */
	err += DrawTitle(win_info);

	/* Draw the legend */
	err += DrawLegend(win_info);

	/* Draw the axis unit labels,  grid lines,  and grid labels */
	err += DrawGridAndAxis(win_info);

	/* Draw the data sets themselves */
	err += DrawData(win_info);
    } else {
      err--;
    }
    return err;
}




static
int DrawTitle(wi)
XGDWin *wi;		/* Window information    */
/*
 * This routine draws the title of the graph centered in
 * the window.  It is spaced down from the top by an amount
 * specified by the constant PADDING.  The font must be
 * fixed width.  The routine returns the height of the
 * title in pixels.
 */
{
    wi->dev_info.xg_text(wi->dev_info.user_state,
			 wi->dev_info.area_w/2,
			 wi->dev_info.axis_pad,
			 PM_STR(wi->pt,"TitleText",wi->interp),
                         T_TOP, 
                         T_TITLE,
                         xg_title0);
    return 0;
}



static
int TransformCompute(wi)
XGDWin *wi;			/* Window information          */
/*
 * This routine figures out how to draw the axis labels and grid lines.
 * Both linear and logarithmic axes are supported.  Axis labels are
 * drawn in engineering notation.  The power of the axes are labeled
 * in the normal axis labeling spots.  The routine also figures
 * out the necessary transformation information for the display
 * of the points (it touches XOrgX, XOrgY, UsrOrgX, UsrOrgY, and
 * UnitsPerPixel).
 */
{
    double bbCenX, bbCenY, bbHalfWidth, bbHalfHeight;
    int idx, maxName, leftWidth;
    char err[MAXBUFSIZE];
    char *XUnitText = PM_STR(wi->pt,"XUnitText",wi->interp);

    /*
     * First,  we figure out the origin in the X window.  Above
     * the space we have the title and the Y axis unit label.
     * To the left of the space we have the Y axis grid labels.
     */

    wi->XOrgX = wi->dev_info.bdr_pad + (7 * wi->dev_info.axis_width)
      + wi->dev_info.bdr_pad;
    wi->XOrgY = wi->dev_info.bdr_pad + wi->dev_info.title_height
      + wi->dev_info.bdr_pad + wi->dev_info.axis_height
	+ wi->dev_info.axis_height/2 + wi->dev_info.bdr_pad;

    /*
     * Now we find the lower right corner.  Below the space we
     * have the X axis grid labels.  To the right of the space we
     * have the X axis unit label and the legend.  We assume the 
     * worst case size for the unit label.
     */

    maxName = 0;
    for (idx = 0;  idx < MAXSETS;  idx++) {
	if (wi->PlotData[idx].list) {
	    int tempSize;

	    tempSize = strlen(wi->PlotData[idx].setName);
	    if (tempSize > maxName) maxName = tempSize;
	}
    }
    /* Worst case size of the X axis label: */
    leftWidth = (strlen(XUnitText)) * wi->dev_info.axis_width;
    if ((maxName*wi->dev_info.axis_width)+wi->dev_info.bdr_pad > leftWidth)
      leftWidth = maxName * wi->dev_info.axis_width + wi->dev_info.bdr_pad;
    
    wi->XOppX = wi->dev_info.area_w - wi->dev_info.bdr_pad - leftWidth;
    wi->XOppY = wi->dev_info.area_h - wi->dev_info.bdr_pad
      - wi->dev_info.axis_height - wi->dev_info.bdr_pad;

    if ((wi->XOrgX >= wi->XOppX) || (wi->XOrgY >= wi->XOppY)) {
	do_error(strcpy(err, "Drawing area is too small\n"));
	return 0;
    }

    /* 
     * We now have a bounding box for the drawing region.
     * Figure out the units per pixel using the data set bounding box.
     */
    wi->XUnitsPerPixel = (wi->hiX - wi->loX)/((double) (wi->XOppX - wi->XOrgX));
    wi->YUnitsPerPixel = (wi->hiY - wi->loY)/((double) (wi->XOppY - wi->XOrgY));

    /*
     * Find origin in user coordinate space.  We keep the center of
     * the original bounding box in the same place.
     */
    bbCenX = (wi->loX + wi->hiX) / 2.0;
    bbCenY = (wi->loY + wi->hiY) / 2.0;
    bbHalfWidth = ((double) (wi->XOppX - wi->XOrgX))/2.0 * wi->XUnitsPerPixel;
    bbHalfHeight = ((double) (wi->XOppY - wi->XOrgY))/2.0 * wi->YUnitsPerPixel;
    wi->UsrOrgX = bbCenX - bbHalfWidth;
    wi->UsrOrgY = bbCenY - bbHalfHeight;
    wi->UsrOppX = bbCenX + bbHalfWidth;
    wi->UsrOppY = bbCenY + bbHalfHeight;

    /*
     * Everything is defined so we can now use the SCREENX and SCREENY
     * transformations.
     */
    return 1;
}

static
int DrawGridAndAxis(wi)
XGDWin *wi;			/* Window information         */
/*
 * This routine draws grid line labels in engineering notation,
 * the grid lines themselves,  and unit labels on the axes.
 */
{
    int expX, expY;		/* Engineering powers */
    int startX;
    int Yspot, Xspot;
    char power[10], value[10], final[MAXBUFSIZE+10];
    double Xincr, Yincr, Xstart, Ystart, Yindex, Xindex, larger;
    XSegment segs[2];
    double initGrid(), stepGrid();
    int tickFlag = PM_BOOL(wi->pt,"Ticks",wi->interp);
    int logXFlag = wi->logxdone; /* computed by Traverse */
    int logYFlag = wi->logydone;
    int negXFlag = PM_BOOL(wi->pt,"NegateX",wi->interp);
    int negYFlag = PM_BOOL(wi->pt,"NegateY",wi->interp);
    char *XUnitText = PM_STR(wi->pt,"XUnitText",wi->interp);
    char *YUnitText = PM_STR(wi->pt,"YUnitText",wi->interp);

    /*
     * Grid display powers are computed by taking the log of
     * the largest numbers and rounding down to the nearest
     * multiple of 3.
     */
    if (logXFlag) {
	expX = 0;
    } else {
	if (fabs(wi->UsrOrgX) > fabs(wi->UsrOppX)) {
	    larger = fabs(wi->UsrOrgX);
	} else {
	    larger = fabs(wi->UsrOppX);
	}
	expX = ((int) floor(nlog10(larger)/3.0)) * 3;
    }
    if (logYFlag) {
	expY = 0;
    } else {
	if (fabs(wi->UsrOrgY) > fabs(wi->UsrOppY)) {
	    larger = fabs(wi->UsrOrgY);
	} else {
	    larger = fabs(wi->UsrOppY);
	}
	expY = ((int) floor(nlog10(larger)/3.0)) * 3;
    }

    /*
     * With the powers computed,  we can draw the axis labels.
     */
    if (expY != 0) {
	(void) strcpy(final, YUnitText);
	(void) strcat(final, " x 10");
	Xspot = wi->dev_info.bdr_pad +
	  ((strlen(YUnitText)+5) * wi->dev_info.axis_width);
	Yspot = wi->dev_info.bdr_pad * 2 + wi->dev_info.title_height +
	  wi->dev_info.axis_height/2;
	wi->dev_info.xg_text(wi->dev_info.user_state,
			     Xspot, Yspot, final, T_RIGHT, T_AXIS, xg_y);
	(void) sprintf(power, "%d", expY);
	wi->dev_info.xg_text(wi->dev_info.user_state,
			     Xspot, Yspot, power, T_LOWERLEFT, T_AXIS,xg_y);
    } else {
	Yspot = wi->dev_info.bdr_pad * 2 + wi->dev_info.title_height;
	wi->dev_info.xg_text(wi->dev_info.user_state,
			     wi->dev_info.bdr_pad, Yspot, YUnitText,
			     T_UPPERLEFT, T_AXIS,xg_y);
    }
	
    startX = wi->dev_info.area_w - wi->dev_info.bdr_pad;
    if (expX != 0) {
	(void) sprintf(power, "%d", expX);
	startX -= (strlen(power) * wi->dev_info.axis_width);
	wi->dev_info.xg_text(wi->dev_info.user_state,
			     startX, wi->XOppY, power,
                             T_LOWERLEFT, T_AXIS, xg_x);
	(void) strcpy(final, XUnitText);
	(void) strcat(final, " x 10");
	wi->dev_info.xg_text(wi->dev_info.user_state,
			     startX, wi->XOppY, final, 
                             T_RIGHT, T_AXIS, xg_x);
    } else {
	wi->dev_info.xg_text(wi->dev_info.user_state,
			     startX, wi->XOppY, XUnitText,
                             T_RIGHT, T_AXIS, xg_x);
    }

    /* 
     * First,  the grid line labels
     */
    Yincr = (wi->dev_info.axis_pad + wi->dev_info.axis_height) *
               wi->YUnitsPerPixel;
    Ystart = initGrid(wi->UsrOrgY, Yincr, logYFlag);
    for (Yindex = Ystart;  Yindex < wi->UsrOppY;  Yindex = stepGrid()) {
	Yspot = SCREENY(wi, Yindex);
	/* Write the axis label */
	WriteValue(value, Yindex, expY, logYFlag,negYFlag);
	wi->dev_info.xg_text(wi->dev_info.user_state,
			     wi->dev_info.bdr_pad +
				     (7 * wi->dev_info.axis_width),
			     Yspot, value, T_RIGHT, T_AXIS,xg_coord);
    }

    Xincr = (wi->dev_info.axis_pad + (wi->dev_info.axis_width * 7)) *
               wi->XUnitsPerPixel;
    Xstart = initGrid(wi->UsrOrgX, Xincr, logXFlag,negXFlag);
    for (Xindex = Xstart;  Xindex < wi->UsrOppX;  Xindex = stepGrid()) {
	Xspot = SCREENX(wi, Xindex);
	/* Write the axis label */
	WriteValue(value, Xindex, expX, logXFlag, negXFlag);
	wi->dev_info.xg_text(wi->dev_info.user_state,
			     Xspot,
			     wi->dev_info.area_h - wi->dev_info.bdr_pad,
			     value, T_BOTTOM, T_AXIS,xg_coord);
    }

    /*
     * Now,  the grid lines or tick marks
     */
    Yincr = (wi->dev_info.axis_pad + wi->dev_info.axis_height) * 
               wi->YUnitsPerPixel;
    Ystart = initGrid(wi->UsrOrgY, Yincr, logYFlag);
    for (Yindex = Ystart;  Yindex < wi->UsrOppY;  Yindex = stepGrid()) {
	Yspot = SCREENY(wi, Yindex);
	/* Draw the grid line or tick mark */
	if (tickFlag) {
	    segs[0].x1 = wi->XOrgX;
	    segs[0].x2 = wi->XOrgX + wi->dev_info.tick_len;
	    segs[1].x1 = wi->XOppX - wi->dev_info.tick_len;
	    segs[1].x2 = wi->XOppX;
	    segs[0].y1 = segs[0].y2 = segs[1].y1 = segs[1].y2 = Yspot;
	} else {
	    segs[0].x1 = wi->XOrgX;  segs[0].x2 = wi->XOppX;
	    segs[0].y1 = segs[0].y2 = Yspot;
	}
	if ((ABS(Yindex) < ZERO_THRES) && !logYFlag) {
	    wi->dev_info.xg_seg(wi->dev_info.user_state,
				 1, segs, PM_INT(wi->pt,"ZeroWidth",wi->interp),
				 L_ZERO, 0, 0, xg_coord);
	    if (tickFlag) {
		wi->dev_info.xg_seg(wi->dev_info.user_state,
				     1, &(segs[1]), 
                                     PM_INT(wi->pt,"ZeroWidth",wi->interp),
				     L_ZERO, 0, 0, xg_coord);
	    }
	} else {
	    wi->dev_info.xg_seg(wi->dev_info.user_state,
				 1, segs, 
                                 PM_INT(wi->pt,"GridSize",wi->interp),
				 L_AXIS, 0, 0, xg_coord);
	    if (tickFlag) {
		wi->dev_info.xg_seg(wi->dev_info.user_state,
				     1, &(segs[1]), 
                                     PM_INT(wi->pt,"GridSize",wi->interp),
				     L_AXIS, 0, 0, xg_coord);
	    }
	}
    }

    Xincr = (wi->dev_info.axis_pad + (wi->dev_info.axis_width * 7)) * 
               wi->XUnitsPerPixel;
    Xstart = initGrid(wi->UsrOrgX, Xincr, logXFlag);
    for (Xindex = Xstart;  Xindex < wi->UsrOppX;  Xindex = stepGrid()) {
	Xspot = SCREENX(wi, Xindex);
	/* Draw the grid line or tick marks */
	if (tickFlag) {
	    segs[0].x1 = segs[0].x2 = segs[1].x1 = segs[1].x2 = Xspot;
	    segs[0].y1 = wi->XOrgY;
	    segs[0].y2 = wi->XOrgY + wi->dev_info.tick_len;
	    segs[1].y1 = wi->XOppY - wi->dev_info.tick_len;
	    segs[1].y2 = wi->XOppY;
	} else {
	    segs[0].x1 = segs[0].x2 = Xspot;
	    segs[0].y1 = wi->XOrgY; segs[0].y2 = wi->XOppY;
	}
	if ((ABS(Xindex) < ZERO_THRES) && !logXFlag) {
	    wi->dev_info.xg_seg(wi->dev_info.user_state,
				 1, segs,
                                 PM_INT(wi->pt,"ZeroWidth",wi->interp),
                                 L_ZERO, 0, 0, xg_coord);
	    if (tickFlag) {
		wi->dev_info.xg_seg(wi->dev_info.user_state,
				    1, &(segs[1]),
                                    PM_INT(wi->pt,"ZeroWidth",wi->interp),
				    L_ZERO, 0, 0, xg_coord);

	    }
	} else {
	    wi->dev_info.xg_seg(wi->dev_info.user_state,
				 1, segs, 
                                 PM_INT(wi->pt,"GridSize",wi->interp),
                                 L_AXIS, 0, 0, xg_coord);
	    if (tickFlag) {
		wi->dev_info.xg_seg(wi->dev_info.user_state,
				     1, &(segs[1]), 
                                     PM_INT(wi->pt,"GridSize",wi->interp),
                                     L_AXIS, 0, 0, xg_coord);
	    }
	}
    }
    /* Check to see if he wants a bounding box */
    if (PM_BOOL(wi->pt,"BoundBox",wi->interp)) {
	XSegment bb[4];

	/* Draw bounding box */
	bb[0].x1 = bb[0].x2 = bb[1].x1 = bb[3].x2 = wi->XOrgX;
	bb[0].y1 = bb[2].y2 = bb[3].y1 = bb[3].y2 = wi->XOrgY;
	bb[1].x2 = bb[2].x1 = bb[2].x2 = bb[3].x1 = wi->XOppX;
	bb[0].y2 = bb[1].y1 = bb[1].y2 = bb[2].y1 = wi->XOppY;
	wi->dev_info.xg_seg(wi->dev_info.user_state,
			     4, bb, PM_INT(wi->pt,"GridSize",wi->interp), 
                             L_AXIS, 0, 0, xg_border);
    }
    return 0;
}

static double gridBase, gridStep, gridJuke[101];
static int gridNJuke, gridCurJuke;

#define ADD_GRID(val)	(gridJuke[gridNJuke++] = log10(val))

double initGrid(low, step, logFlag)
double low;			/* desired low value          */
double step;			/* desired step (user coords) */
int logFlag;			/* is axis logarithmic?       */
{
    double ratio, x;
    double RoundUp(), stepGrid();

    gridNJuke = gridCurJuke = 0;
    gridJuke[gridNJuke++] = 0.0;

    if (logFlag) {
	ratio = pow(10.0, step);
	gridBase = floor(low);
	gridStep = ceil(step);
	if (ratio <= 3.0) {
	    if (ratio > 2.0) {
		ADD_GRID(3.0);
	    } else if (ratio > 1.333) {
		ADD_GRID(2.0);	ADD_GRID(5.0);
	    } else if (ratio > 1.25) {
		ADD_GRID(1.5);	ADD_GRID(2.0);	ADD_GRID(3.0);
		ADD_GRID(5.0);	ADD_GRID(7.0);
	    } else {
		for (x = 1.0; x < 10.0 && (x+.5)/(x+.4) >= ratio; x += .5) {
		    ADD_GRID(x + .1);	ADD_GRID(x + .2);
		    ADD_GRID(x + .3);	ADD_GRID(x + .4);
		    ADD_GRID(x + .5);
		}
		if (floor(x) != x) ADD_GRID(x += .5);
		for ( ; x < 10.0 && (x+1.0)/(x+.5) >= ratio; x += 1.0) {
		    ADD_GRID(x + .5);	ADD_GRID(x + 1.0);
		}
		for ( ; x < 10.0 && (x+1.0)/x >= ratio; x += 1.0) {
		    ADD_GRID(x + 1.0);
		}
		if (x == 7.0) {
		    gridNJuke--;
		    x = 6.0;
		}
		if (x < 7.0) {
		    ADD_GRID(x + 2.0);
		}
		if (x == 10.0) gridNJuke--;
	    }
	    x = low - gridBase;
	    for (gridCurJuke = -1; x >= gridJuke[gridCurJuke+1]; gridCurJuke++){
	    }
	}
    } else {
	gridStep = RoundUp(step);
	gridBase = floor(low / gridStep) * gridStep;
    }
    return(stepGrid());
}

double stepGrid()
{
    if (++gridCurJuke >= gridNJuke) {
	gridCurJuke = 0;
	gridBase += gridStep;
    }
    return(gridBase + gridJuke[gridCurJuke]);
}

double RoundUp(val)
double val;			/* Value */
/*
 * This routine rounds up the given positive number such that
 * it is some power of ten times either 1, 2, or 5.  It is
 * used to find increments for grid lines.
 */
{
    int exponent, idx;

    exponent = (int) floor(nlog10(val));
    if (exponent < 0) {
	for (idx = exponent;  idx < 0; idx++) {
	    val *= 10.0;
	}
    } else {
	for (idx = 0;  idx < exponent; idx++) {
	    val /= 10.0;
	}
    }
    if (val > 5.0) val = 10.0;
    else if (val > 2.0) val = 5.0;
    else if (val > 1.0) val = 2.0;
    else val = 1.0;
    if (exponent < 0) {
	for (idx = exponent;  idx < 0;  idx++) {
	    val /= 10.0;
	}
    } else {
	for (idx = 0;  idx < exponent;  idx++) {
	    val *= 10.0;
	}
    }
    return val;
}

static
void WriteValue(str, val, expv, logFlag, negFlag)
char *str;			/* String to write into */
double val;			/* Value to print       */
int expv;			/* Exponent             */
int logFlag;			/* Is this a log axis?  */
int negFlag;			/* flip sign on this axis?  */
/*
 * Writes the value provided into the string in a fixed format
 * consisting of seven characters.  The format is:
 *   -ddd.dd
 */
{
    int idx,sign;

    sign = (negFlag ? -1 : 1);
    if (logFlag) {
	if (val == floor(val)) {
	    (void) sprintf(str, "%.0e", sign * pow(10.0, val));
	} else {
	    (void) sprintf(str, "%.2g", sign * pow(10.0, val - floor(val)));
	}
    } else {
	if (expv < 0) {
	    for (idx = expv;  idx < 0;  idx++) {
		val *= 10.0;
	    }
	} else {
	    for (idx = 0;  idx < expv;  idx++) {
		val /= 10.0;
	    }
	}
	(void) sprintf(str, "%.2f", (val != 0.0) ? sign * val : 0.0);
    }
}
	    
    
#define LEFT_CODE	0x01
#define RIGHT_CODE	0x02
#define BOTTOM_CODE	0x04
#define TOP_CODE	0x08

/* Clipping algorithm from Neumann and Sproull by Cohen and Sutherland */
#define C_CODE(xval, yval, rtn) \
rtn = 0; \
if ((xval) < wi->UsrOrgX) rtn = LEFT_CODE; \
else if ((xval) > wi->UsrOppX) rtn = RIGHT_CODE; \
if ((yval) < wi->UsrOrgY) rtn |= BOTTOM_CODE; \
else if ((yval) > wi->UsrOppY) rtn |= TOP_CODE

static
int DrawData(wi)
XGDWin *wi;
/*
 * This routine draws the data sets themselves using the macros
 * for translating coordinates.
 */
{
    double sx1, sy1, sx2, sy2, tx, ty;
    int idx, subindex;
    int code1, code2, cd, mark_inside;
    int X_idx;
    XSegment *ptr;
    PointList *thisList;
    int markFlag, pixelMarks, bigPixel, colorMark;
    int noLines = PM_BOOL(wi->pt,"NoLines",wi->interp);
    int lineWidth = PM_INT(wi->pt,"LineWidth",wi->interp);

    tx = ty = 0.0;

    set_mark_flags(&markFlag, &pixelMarks, &bigPixel, &colorMark,
                   wi->interp,wi->pt);
    for (idx = 0;  idx < MAXSETS;  idx++) {
	thisList = wi->PlotData[idx].list;
	while (thisList) {
	    X_idx = 0;
	    for (subindex = 0;  subindex < thisList->numPoints-1;  subindex++) {
		/* Put segment in (sx1,sy1) (sx2,sy2) */
		sx1 = thisList->xvec[subindex];
		sy1 = thisList->yvec[subindex];
		sx2 = thisList->xvec[subindex+1];
		sy2 = thisList->yvec[subindex+1];
		/* Now clip to current window boundary */
		C_CODE(sx1, sy1, code1);
		C_CODE(sx2, sy2, code2);
		mark_inside = (code1 == 0);
		while (code1 || code2) {
		    if (code1 & code2) break;
		    cd = (code1 ? code1 : code2);
		    if (cd & LEFT_CODE) {	/* Crosses left edge */
			ty = sy1 + (sy2 - sy1) * (wi->UsrOrgX - sx1) / 
                                    (sx2 - sx1);
			tx = wi->UsrOrgX;
		    } else if (cd & RIGHT_CODE) { /* Crosses right edge */
			ty = sy1 + (sy2 - sy1) * (wi->UsrOppX - sx1) / 
                                    (sx2 - sx1);
			tx = wi->UsrOppX;
		    } else if (cd & BOTTOM_CODE) { /* Crosses bottom edge */
			tx = sx1 + (sx2 - sx1) * (wi->UsrOrgY - sy1) / 
                                    (sy2 - sy1);
			ty = wi->UsrOrgY;
		    } else if (cd & TOP_CODE) { /* Crosses top edge */
			tx = sx1 + (sx2 - sx1) * (wi->UsrOppY - sy1) / 
                                    (sy2 - sy1);
			ty = wi->UsrOppY;
		    }
		    if (cd == code1) {
			sx1 = tx;  sy1 = ty;
			C_CODE(sx1, sy1, code1);
		    } else {
			sx2 = tx;  sy2 = ty;
			C_CODE(sx2, sy2, code2);
		    }
		}
		if (!code1 && !code2) {
		    /* Add segment to list */
		    wi->Xsegs[X_idx].x1 = SCREENX(wi, sx1);
		    wi->Xsegs[X_idx].y1 = SCREENY(wi, sy1);
		    wi->Xsegs[X_idx].x2 = SCREENX(wi, sx2);
		    wi->Xsegs[X_idx].y2 = SCREENY(wi, sy2);
		    X_idx++;
		}

		/* Draw markers if requested and they are in drawing region */
		if (markFlag && mark_inside) {
		    if (pixelMarks) {
			if (bigPixel) {
			    wi->dev_info.xg_dot(wi->dev_info.user_state,
						wi->Xsegs[X_idx-1].x1,
                                                wi->Xsegs[X_idx-1].y1,
						P_DOT, 0, idx % MAXATTR,
                                                idx/*xg_series_*/);
			} else {
			    wi->dev_info.xg_dot(wi->dev_info.user_state,
						wi->Xsegs[X_idx-1].x1,
                                                wi->Xsegs[X_idx-1].y1,
						P_PIXEL, 0, PIXVALUE(idx),
                                                idx/*xg_series_*/);
			}
		    } else {
			/* Distinctive markers */
			wi->dev_info.xg_dot(wi->dev_info.user_state,
					    wi->Xsegs[X_idx-1].x1,
                                            wi->Xsegs[X_idx-1].y1,
					    P_MARK, MARKSTYLE(idx),
					    PIXVALUE(idx),
                                            idx/*xg_series_*/);
		    }
		}

		/* Draw bar elements if requested */
		if (PM_BOOL(wi->pt,"BarGraph",wi->interp)) {
		    int barPixels, baseSpot;
		    XSegment line;

		    barPixels = (int) ((PM_DBL(wi->pt,"BarWidth",wi->interp)/
                                        wi->XUnitsPerPixel) + 0.5);
		    if (barPixels <= 0) barPixels = 1;
		    baseSpot = SCREENY(wi, PM_DBL(wi->pt,"BarBase",wi->interp));
		    line.x1 = line.x2 = wi->Xsegs[X_idx-1].x1;
		    line.y1 = baseSpot;  
                    line.y2 = wi->Xsegs[X_idx-1].y1;
		    wi->dev_info.xg_seg(wi->dev_info.user_state,
					1, &line, barPixels, L_VAR,
					LINESTYLE(idx), PIXVALUE(idx),
                                        idx/*xg_series_*/);
		}
	    }
	    /* Handle last marker */
	    if (markFlag && (thisList->numPoints > 0)) {
		C_CODE(thisList->xvec[thisList->numPoints-1],
		       thisList->yvec[thisList->numPoints-1],
		       mark_inside);
		if (mark_inside == 0) {
		    if (pixelMarks) {
			if (bigPixel) {
			    wi->dev_info.xg_dot(wi->dev_info.user_state,
						wi->Xsegs[X_idx-1].x2,
                                                wi->Xsegs[X_idx-1].y2,
						P_DOT, 0, idx % MAXATTR,
                                                idx/*xg_series_*/);
			} else {
			    wi->dev_info.xg_dot(wi->dev_info.user_state,
						wi->Xsegs[X_idx-1].x2, 
                                                wi->Xsegs[X_idx-1].y2,
						P_PIXEL, 0, PIXVALUE(idx),
                                                idx/*xg_series_*/);
			}
		    } else {
			/* Distinctive markers */
			wi->dev_info.xg_dot(wi->dev_info.user_state,
					    wi->Xsegs[X_idx-1].x2, 
                                            wi->Xsegs[X_idx-1].y2,
					    P_MARK, MARKSTYLE(idx),
					    PIXVALUE(idx),
                                            idx/*xg_series_*/);
		    }
		}
	    }
	    /* Handle last bar */
	    if ((thisList->numPoints > 0) &&
                 PM_BOOL(wi->pt,"BarGraph",wi->interp)) {
		int barPixels, baseSpot;
		XSegment line;

		barPixels = (int) ((PM_DBL(wi->pt,"BarWidth",wi->interp)/
                                    wi->XUnitsPerPixel) + 0.5);
		if (barPixels <= 0) barPixels = 1;
		baseSpot = SCREENY(wi, PM_DBL(wi->pt,"BarBase",wi->interp));
		line.x1 = line.x2 = wi->Xsegs[X_idx-1].x2;
		line.y1 = baseSpot; 
                line.y2 = wi->Xsegs[X_idx-1].y2;
		wi->dev_info.xg_seg(wi->dev_info.user_state,
				    1, &line, barPixels, L_VAR,
				    LINESTYLE(idx), PIXVALUE(idx),
                                    idx/*xg_series_*/);
	    }

	    /* Draw segments in chunks of max_segs or less. */
	    if (thisList->numPoints > 0 && (!noLines) && (X_idx > 0)) {
		ptr = wi->Xsegs;
		while (X_idx > wi->dev_info.max_segs) {
		    wi->dev_info.xg_seg(wi->dev_info.user_state,
					wi->dev_info.max_segs, ptr,
					lineWidth, L_VAR,
					LINESTYLE(idx), PIXVALUE(idx),
                                        idx/*xg_series_*/);
		    ptr += wi->dev_info.max_segs;
		    X_idx -= wi->dev_info.max_segs;
		}
		wi->dev_info.xg_seg(wi->dev_info.user_state,
				    X_idx, ptr,
				    lineWidth, L_VAR,
				    LINESTYLE(idx), PIXVALUE(idx),
                                    idx/*xg_series_*/);
	    }
	    /* Next subset */
	    thisList = thisList->next;
	}
    }
    return 0;
}


static
int DrawLegend(wi)
XGDWin *wi;
/*
 * This draws a legend of the data sets displayed.  Only those that
 * will fit are drawn.
 */
{
    int idx, spot, lineLen, oneLen;
    XSegment leg_line;
    int markFlag, pixelMarks, bigPixel, colorMark;

    set_mark_flags(&markFlag, &pixelMarks, &bigPixel, &colorMark,wi->interp,
                   wi->pt);
    spot = wi->XOrgY;
    lineLen = 0;
    /* First pass draws the text */
    for (idx = 0;  idx < MAXSETS;  idx++) {
	if ((wi->PlotData[idx].list) &&
	    (spot + wi->dev_info.axis_height + 2 < wi->XOppY))
	  {
	      /* Meets the criteria */
	      oneLen = strlen(wi->PlotData[idx].setName);
	      if (oneLen > lineLen) lineLen = oneLen;
	      wi->dev_info.xg_text(wi->dev_info.user_state,
				   wi->XOppX + wi->dev_info.bdr_pad,
				   spot+2,
				   wi->PlotData[idx].setName,
				   T_UPPERLEFT, T_AXIS, idx/*xg_series_*/);
	      spot += 2 + wi->dev_info.axis_height + wi->dev_info.bdr_pad;
	  }
    }
    lineLen = lineLen * wi->dev_info.axis_width;
    leg_line.x1 = wi->XOppX + wi->dev_info.bdr_pad;
    leg_line.x2 = leg_line.x1 + lineLen;
    spot = wi->XOrgY;
    /* second pass draws the lines */
    for (idx = 0;  idx < MAXSETS;  idx++) {
	if ((wi->PlotData[idx].list) &&
	    (spot + wi->dev_info.axis_height + 2 < wi->XOppY))
	  {
	      leg_line.y1 = leg_line.y2 = spot - wi->dev_info.legend_pad;
	      wi->dev_info.xg_seg(wi->dev_info.user_state,
				  1, &leg_line, 1, L_VAR,
				  LINESTYLE(idx), PIXVALUE(idx),
                                  idx/*xg_series_*/);
	      if (markFlag && !pixelMarks) {
		  wi->dev_info.xg_dot(wi->dev_info.user_state,
				      leg_line.x1, leg_line.y1,
				      P_MARK, MARKSTYLE(idx), PIXVALUE(idx),
                                      idx/*xg_series_*/);
		  
	      }
	      spot += 2 + wi->dev_info.axis_height + wi->dev_info.bdr_pad;
	  }
    }
    return 0;
}



static 
void set_mark_flags(markFlag, pixelMarks, bigPixel, colorMark,interp,pt)
int *markFlag;
int *pixelMarks;
int *bigPixel;
int *colorMark;
Tcl_Interp *interp;
st_table *pt;
/*
 * Determines the values of the old boolean flags based on the
 * new values in the parameters database.
 */
{
    *markFlag = 0;  *pixelMarks = 0;  *colorMark = 0;  *bigPixel = 0;
    if (PM_BOOL(pt,"Markers",interp)) {
	*markFlag = 1;  *pixelMarks = 0;  *colorMark = 0;
    }
    if (PM_BOOL(pt,"PixelMarkers",interp)) {
	*markFlag = 1;  *pixelMarks = 1;  *bigPixel = 0;
    }
    if (PM_BOOL(pt,"LargePixels",interp)) {
	*markFlag = 1;  *pixelMarks = 1;  *bigPixel = 1;
    }
    if (PM_BOOL(pt,"StyleMarkers",interp)) {
	*markFlag = 1;  *pixelMarks = 0;  *colorMark = 1;
    }
}



#define RND(val)	((int) ((val) + 0.5))



#define ERR_MSG_SIZE	2048



/* returns default values if unsuccessful.
 * parses specs of form WxH+X+Y, WXH, +X+Y 
*/
static
void parse_geometry(geo,spec)
Geom *geo;
char *spec;
{
  Geom tmp;
  if (sscanf(spec,"%dx%d+%d+%d",&(tmp.w),&(tmp.h),&(tmp.x),&(tmp.y)) == 4) {
    *geo = tmp;
    return ;
  }
  if (sscanf(spec,"%dx%d",&(tmp.w),&(tmp.h)) == 2) {
    *geo = tmp;
    geo->x = 20;
    geo->y = 20;
    return ;
  }
  if (sscanf(spec,"+%d+%d",&(tmp.x),&(tmp.y)) == 2) {
    *geo = tmp;
    geo->w = geo->h = NORMSIZE;
    return ;
  }
  geo->x = geo->y = 20;
  geo->w = geo->h = NORMSIZE;
}

/*
 * returns 0 if ok and other if error in data reading of any sort.
 */
static
int Tkxg_LoadFiles(wi,interp,argc,argv)
XGDWin *wi;
Tcl_Interp *interp;
int argc;
char *argv[];
{

    FILE *strm;
    int idx, maxitems;
    int errs = 0;
    maxitems = 0;
   

    /* Parse the argument list looking for input files. 
     * it should be nothing but files.
     */
    ParseArgs(argc, argv, 2,interp,wi->pt,wi);

    /* Read the data into the data sets */
    wi->llx = wi->lly = MAXFLOAT;
    wi->urx = wi->ury = -MAXFLOAT;
    for (idx = 0;  idx < wi->numFiles;  idx++) {
/* stdin is not supported as a file any more. write a different entry
 * for loadfiles if you need it.
 *	strm = strcmp (inFileNames[idx], "-") ? fopen(inFileNames[idx], "r") :
 *	       stdin;
 */
	strm = fopen(wi->inFileNames[idx], "r");
	if (strm==NULL) {
	    (void) fprintf(stderr, "Warning:  cannot open file `%s'\n",
			   wi->inFileNames[idx]);
            errs++;
            break;
	} else {
	    maxitems = ReadData(strm, wi->inFileNames[idx],wi->pt,wi);
	    if (strm != stdin) (void) fclose(strm);
	    if (maxitems < 0) {
		errs++;
                break;
	    }
	}
    }
/*
 *    if (!(wi->numFiles)) {
 *	if ((maxitems = ReadData(stdin, (char *) 0)) < 0) {
 *	    errs++;
 *	}
 *   }
 */
    if (errs || wi->numFiles == 0) {
	Tcl_AppendResult(interp,"Problems found with input data.\n",
        (char *)NULL);
	return 1;
    }

    wi->Xsegs = (XSegment *) malloc((unsigned) (maxitems * sizeof(XSegment)));
    return 0;
}

static 
int Tkxg_Plot(wi,interp)
XGDWin *wi;
Tcl_Interp *interp;
{
    int flags;
    Geom geom;

    /* Logarithmic and bounding box computation.
     */
    wi->llx = wi->lly = MAXFLOAT;
    wi->urx = wi->ury = -MAXFLOAT;
    flags = 0;
    if (PM_BOOL(wi->pt,"LogX",interp)) {
        flags |= LOG_X;
    }
    if (PM_BOOL(wi->pt,"LogY",interp) ) {
        flags |= LOG_Y;
    }
    if (Traverse(flags,wi) != TCL_OK) { /* never fails now */
	return TCL_ERROR;
    }

    /* Nasty hack here for bar graphs */
    if (PM_BOOL(wi->pt,"BarGraph",interp)) {
	double base;

	wi->llx -= PM_DBL(wi->pt,"BarWidth",interp);
	wi->urx += PM_DBL(wi->pt,"BarWidth",interp);
	base = PM_DBL(wi->pt,"BarBase",interp);
	if (base < wi->lly) wi->lly = base;
	if (base > wi->ury) wi->ury = base;
    }

    Tkxg_PlotSize(wi,
            PM_DBL(wi->pt,"XLowLimit",interp),
            PM_DBL(wi->pt,"YLowLimit",interp),
            PM_DBL(wi->pt,"XHighLimit",interp), 
            PM_DBL(wi->pt,"YHighLimit",interp),
            1.0,interp);

    parse_geometry(&geom,PM_STR(wi->pt,"Geometry",interp));
    wi->dev_info.area_w = geom.w;
    wi->dev_info.area_h = geom.h;
    init_X(wi->dev_info.user_state,wi->pt);
    reset_tk(&(wi->dev_info),interp,wi->pt);

    if (DrawWindow(wi)!=0) return TCL_ERROR;
    return TCL_OK;
}

/*
 * This function returns a string which is good until the next call of
 * this function. The caller must not free the string returned, but
 * can change it's content up to the first NULL character.
 */
static
char *OptionValueString(sp,wi,interp,series)
xg_configspec *sp;
XGDWin *wi;
Tcl_Interp *interp;
int series;
{
    static char buf[256];
    static char *tmps;
    
    buf[0]= '{';
    buf[1]= '}';
    tmps = buf;
    switch(sp->type) {
    case TK_CONFIG_FONT:
        tmps = PM_TKFONT(wi->pt,sp->dbName,interp);
        break;
    case TK_CONFIG_COLOR:
        tmps = PM_TKCOLOR(wi->pt,sp->dbName,interp);
        break;
    case TK_CONFIG_BITMAP:
        tmps = PM_STR(wi->pt,sp->dbName,interp);
        break;
    case TK_CONFIG_STRING:
        tmps = PM_STR(wi->pt,sp->dbName,interp);
        break;
    case TK_CONFIG_DOUBLE:
        sprintf(buf,"%.18g",PM_DBL(wi->pt,sp->dbName,interp));
        break;
    case TK_CONFIG_PIXELS:
        sprintf(buf,"%d",PM_INT(wi->pt,sp->dbName,interp));
        break;
    case TK_CONFIG_BOOLEAN:
        sprintf(buf,"%d",PM_BOOL(wi->pt,sp->dbName,interp));
        break;
    case XG_CONFIG_MANY:
        if (series<0 || series >= wi->nseries) {
            tmps = "SERIES RANGE ERROR";
            break;
        }
        sprintf(buf,sp->dbName,series);
        switch (sp->argvName[7]) { 
        /* this is gross to be depending that XG <-> -series */
        case 'l':
        case 's':
        case 'm':
        case 'c':
            tmps = PM_STR(wi->pt,buf,interp);
            break;
        case 'w':
            sprintf(buf,"%d",PM_INT(wi->pt,buf,interp));
            break;
        default:
            break;
        }
    default:
        break;
    }
    return tmps;
}

static
void WriteConfiguration(sp,wi,interp,list)
xg_configspec *sp;
XGDWin *wi;
Tcl_Interp *interp;
int list;
{
#define SETRESULT(s) \
  if (list) Tcl_AppendElement(interp,s); \
  else Tcl_AppendResult(interp,s,(char *)NULL)

    char *tmps, *longs;
    static char buf[1024];
    
    if (wi==NULL || interp == NULL || sp == NULL) {
        return;
    }
    switch(sp->type) {
    case TK_CONFIG_FONT:
    case TK_CONFIG_COLOR:
    case TK_CONFIG_BITMAP:
        tmps = OptionValueString(sp,wi,interp,-1);
        sprintf(buf,"%s %s %s {%s} {%s}",
                sp->argvName, sp->dbName, sp->dbClass, sp->dbInit,tmps);
        SETRESULT(buf);
        break;
    case TK_CONFIG_STRING:
        tmps = OptionValueString(sp,wi,interp,-1);
        sprintf(buf,"%s %s %s {%s} {",
                sp->argvName, sp->dbName, sp->dbClass, sp->dbInit);
        longs = malloc(strlen(buf)+strlen(tmps)+2); /*}\0*/
        if (longs ==NULL) {
            break; /* too big? */
        }
        sprintf(longs,"%s%s}",buf,tmps);
        SETRESULT(longs);
        free(longs);
        break;
    case TK_CONFIG_DOUBLE:
    case TK_CONFIG_BOOLEAN:
    case TK_CONFIG_PIXELS:
        tmps = OptionValueString(sp,wi,interp,-1);
        sprintf(buf,"%s %s %s %s %s",
                sp->argvName, sp->dbName, sp->dbClass, sp->dbInit,tmps);
        SETRESULT(buf);
        break;
    case XG_CONFIG_MANY:
        /* either return empty or return a list for the last element */
        sprintf(buf,"%s %s %s {%s} {}",
                sp->argvName, sp->dbName, sp->dbClass, sp->dbInit);
        SETRESULT(buf);
        break;
    case TK_CONFIG_END:
        break;
    default:
        fprintf(stderr,"tkxgraph: UNKNOWN OPTION TYPE\n");
        break;
    }
#undef SETRESULT
}

static
void WriteConfigurationList(sp,wi,interp)
xg_configspec *sp;
XGDWin *wi;
Tcl_Interp *interp;
{
    if (sp==NULL) {
        return;
    }
    for (sp = tkspecs; sp->type != TK_CONFIG_END; sp++) {
        WriteConfiguration(sp,wi,interp,1);
    }
}

/* returns pointer to spec matching name, or to the end spec if no match */
static
xg_configspec *Tkxg_FindSpec(sp,name)
xg_configspec *sp;
char *name;
{
    if (sp==NULL || name == NULL) {
        return NULL;
    }
    while(sp->type != TK_CONFIG_END && (strcmp(name,sp->argvName)!=0)) {
        sp++;
    }
    return sp;
}

static
int Tkxg_Configure (wi,interp,argc,argv)
XGDWin *wi;
Tcl_Interp *interp;
int argc;
char *argv[];
{
    xg_configspec *sp;
    int idx;

    if (argv[1][1]== '\0') {
        Tcl_AppendResult(interp,argv[1]," ambiguous ",
                 "option of ",argv[0],": need configure or cget",
                 (char *)NULL);
        return TCL_ERROR;
    }
    if (argc==2) { 
        /* configure no args */
        if (argv[1][1]=='o') {
	    WriteConfigurationList(tkspecs,wi,interp);
            return TCL_OK;
        }
    }
    if (argc == 3) {
	sp = Tkxg_FindSpec(tkspecs,argv[2]);
        if ( sp->type == TK_CONFIG_END ) {
            Tcl_AppendResult(interp,argv[2]," unknown or unsupported ",
                     "option of ",argv[0]," ",argv[1],(char *)NULL);
            return TCL_ERROR;
        }
        if (argv[1][1]=='g') {
            /* cget one arg */
            Tcl_AppendResult(interp,
                    OptionValueString(sp,wi,interp,-1),
                    (char *)NULL);
            return TCL_OK;
        }
        if (argv[1][1]=='o') {
            /* configure one arg */
            WriteConfiguration(sp,wi,interp,0);
            return TCL_OK;
        }
        Tcl_AppendResult(interp,argv[1]," unknown option of ",
                argv[0],(char *)NULL);
        return TCL_ERROR;
    }

    /* two arg cget , assuming series <==> XG_CONFIG_MANY. */
    if (argc == 4 && argv[1][1]=='g' && strncmp("-series",argv[2],7)==0 ) {
        int status,series;
        sp = Tkxg_FindSpec(tkspecs,argv[2]);
        if ( sp->type == TK_CONFIG_END ) {
            Tcl_AppendResult(interp,argv[2]," unknown or unsupported ",
                     "option of ",argv[0]," ",argv[1],(char *)NULL);
            return TCL_ERROR;
        }
        status = Tcl_GetInt(interp,argv[3],&series);
        if (status != TCL_OK) {
            Tcl_AppendResult(interp,"Unable to parse series number ",argv[3],
                     " of ",argv[0]," ",argv[1]," ",argv[2],(char *)NULL);
            return TCL_ERROR;
        }
        /* cget two arg */
        Tcl_AppendResult(interp,
                OptionValueString(sp,wi,interp,series),
                (char *)NULL);
        return TCL_OK;
    }

    /* check for files */
    for (idx = 2; idx < argc; idx++) {
        if (strcmp("-f",argv[idx])==0 || strcmp("-file",argv[idx])==0) {
            Tcl_AppendResult(interp,argv[0]," ",argv[1],": files cannot",
                     " be added by configure ",argv[idx],(char *)NULL);
            if (idx+1 < argc) {
                Tcl_AppendResult(interp,argv[idx+1],(char *)NULL);
            }
            return TCL_ERROR;
        }
    }

    /* Parse the argument list to set non-file options */
    if (ParseArgs(argc-2, argv+2, 1,interp,wi->pt,wi) != TCL_OK) {
        Tcl_AppendResult(interp,argv[0]," ",argv[1]," error.",(char *)NULL);
        return TCL_ERROR;
    }
    return TCL_OK;
}

static
int Tkxg_maxsets(interp,wi)
Tcl_Interp *interp;
XGDWin *wi;
{
    char tmps[40];
    if (interp==NULL || wi==NULL || wi->nseries <0 || wi->nseries >MAXSETS) {
        return TCL_ERROR;
    }
    sprintf(tmps,"%d",wi->nseries);
    Tcl_AppendResult(interp,tmps,(char *)NULL);
    return TCL_OK;
}

static
int Tkxg_param(wi,interp,argc,argv)
XGDWin *wi;
Tcl_Interp *interp;
int argc;
char *argv[];
{
    params val;
    if (argc==2) {
        param_dump(wi->pt);
        return TCL_OK;
    }
    /* here we should have a resource lookup */
    if (argc==3) {

        /* Get attributes out parameters database */
        (void) param_get(wi->pt,argv[2], &val,interp);
        switch (val.type) {
/*
        case INT:
            Tcl_AppendResult(interp,"INT ",val.text_form,(char *)NULL);
            break;
        case BOOL:
            Tcl_AppendResult(interp,"BOOL ",val.text_form,(char *)NULL);
          break;
        case DBL:
            Tcl_AppendResult(interp,"DBL ",val.text_form,(char *)NULL);
            break;
*/
        case PIXEL:
            Tcl_AppendResult(interp,"PIXEL ",val.pixv.tkcolorname,(char *)NULL);
            break;
        case STR:
            Tcl_AppendResult(interp,"STR {",val.strv.value,"}",(char *)NULL);
            break;
        case FONT:
            Tcl_AppendResult(interp,"FONT {",val.fontv.tkfontname,"}",(char *)NULL);
            break;
        case STYLE:
            Tcl_AppendResult(interp,"STYLE {",val.stylev.tkbitname,"}",(char *)NULL);
            break;
        default:
            Tcl_AppendResult(interp,"Unknown resource.",(char *)NULL);
            return TCL_ERROR;
        }
        return TCL_OK;
    }
    Tcl_AppendResult(interp,argv[0]," ",argv[1]," wants 1 resource name",
            (char *)NULL);
    return TCL_ERROR;
}

int Tkxg_Update (cdata,interp,argc,argv)
ClientData cdata;
Tcl_Interp *interp;
int argc;
char *argv[];
/*
 * This is the xgraphN command. It distributes to one function
 * that destroys the graph and the command, another
 * that parses new args or returns options, and a third that plots.
 * syntax:
 * xgraphN destroy
 * xgraphN plot
 * xgraphN configure
 * xgraphN cget
 * xgraphN resources (private)
 */
{
    char destroycmd[80];
    XGDWin *wi;
    wi = (XGDWin *)cdata;
    if (wi==NULL || wi->magic != TKXGMAGIC) {
        Tcl_AppendResult(interp,"Tkxg_Update called with unknown pointer",
                 (char *)NULL);
        return TCL_ERROR;
    }
    if (argc <2) {
        Tcl_AppendResult(interp,argv[0]," : <destroy,plot,configure,maxsets>",
                 (char *)NULL);
        return TCL_ERROR;
    }
    switch (argv[1][0]) {
    case 'd':
        sprintf(destroycmd,"rename %s {}",wi->cmdname);
        Tcl_GlobalEval(interp,destroycmd);
        return TCL_OK;
    case 'p':
        return Tkxg_Plot(wi,interp);
    case 'c':
        return Tkxg_Configure(wi,interp,argc,argv);
    case 'm':
        return Tkxg_maxsets(interp,wi);
    case 'r':
        return Tkxg_param(wi,interp,argc,argv);
    default:
        break;
    }
    Tcl_AppendResult(interp,argv[0]," : <destroy,plot,configure,maxsets>",
              (char *)NULL);
    return TCL_ERROR;
}

static
void Tkxg_Destroy(cdata)
ClientData cdata;
{
    int i;
    XGDWin *wi;
    wi = (XGDWin *)cdata;
    if (wi==NULL || wi->magic != TKXGMAGIC) {
        return;
    }
    /* this needs to be filled in */
    if (wi->pt) {
        param_destroy(wi->pt);
        wi->pt = NULL;
    }
    if (wi->Xsegs) {
        free(wi->Xsegs);
        wi->Xsegs = NULL;
    }
    for (i=0; i < MAXSETS; i++) {
        if (wi->inFileNames[i] != NULL) {
            free(wi->inFileNames[i]);
            wi->inFileNames[i] = NULL;
        }
        if (wi->PlotData[i].list != NULL) {
            if (wi->PlotData[i].list->xvec != NULL) {
                free(wi->PlotData[i].list->xvec);
                wi->PlotData[i].list->xvec = NULL;
            }
            if (wi->PlotData[i].list->yvec != NULL) {
                free(wi->PlotData[i].list->yvec);
                wi->PlotData[i].list->yvec = NULL;
            }
            free(wi->PlotData[i].list);
            wi->PlotData[i].list = NULL;
        }
    }
    if (wi->cmdname != NULL) {
        free(wi->cmdname);
        wi->cmdname = NULL;
    }
    wi->dev_info.xg_del(wi->dev_info.user_state);
    wi->magic = TKXGBADMAGIC;
    free(wi);
}


int Tkxg_Create (cdata,interp,argc,argv)
ClientData cdata;
Tcl_Interp *interp;
int argc;
char *argv[];
/*
 * This sets up the hard-wired defaults and ignores the X defaults.
 * The command line format is:  tkxgraph options, which must include
 * a file.
 */
{
    FILE *strm;
    int idx, maxitems;
    XGDWin *wi;
    char newname[50];
    char *filelist[MAXSETS];

    (void)cdata;
    maxitems = 0;
    for (idx = 1; idx < argc-1; idx++) {
        if (strcmp("-f",argv[idx])==0 || strcmp("-file",argv[idx])==0) {
            if (argv[idx+1][0] != '\0' && maxitems < MAXSETS) {
                filelist[maxitems++] = argv[idx+1];
                strm = fopen(argv[idx+1],"r");
                if (strm == NULL) {
                    Tcl_AppendResult(interp,argv[0],": unreadable ",
                             argv[idx+1],(char *)NULL);
                    return TCL_ERROR;
                }
                fclose(strm);
            }
            idx++; /* don't bother checking filename for -f */
        }
    }

    if (!maxitems) {
        Tcl_AppendResult(interp,argv[0],": no input files found ",
                         (char *)NULL);
        return TCL_ERROR;
    }

    /* call newwindow to create data structures. */
    wi = NewWindow(interp);

    /* parse files, all of argv should be valid files  */
    if (Tkxg_LoadFiles(wi,interp,maxitems,filelist)) {
        Tkxg_Destroy((ClientData)wi);
        Tcl_AppendResult(interp,argv[0],": unreadable file",(char *)NULL);
        return TCL_ERROR;
    }
    for (idx = 0;  idx < MAXSETS;  idx++) {
        if( wi->PlotData[idx].list != NULL) {
            wi->nseries++;
        }
    }

    /* Parse the argument list to set non-file options */
    if (ParseArgs(argc-1, argv+1, 1,interp,wi->pt,wi) != TCL_OK) {
        Tkxg_Destroy((ClientData)wi);
        Tcl_AppendResult(interp,argv[0],": unreadable file options",
                 (char *)NULL);
        return TCL_ERROR;
    }

    /* define new graph command */
    sprintf(newname,"%s%u",TKBASENAME,name_counter++);
    wi->cmdname = STRDUP(newname);
    Tcl_CreateCommand(interp,newname,Tkxg_Update,
                       (ClientData)wi,Tkxg_Destroy);
    Tcl_AppendResult(interp,newname,(char *)NULL);
    return TCL_OK;
}

static char initScript[] =
"proc tkxgInit {} {\n\
    global tk_library tk_version tk_patchLevel env errorInfo tcl_rcFileName\n\
    rename tkxgInit {}\n\
    set errors \"\"\n\
    if {![info exists tk_library]} {\n\
        set tk_library .\n\
    }\n\
    set dirs {}\n\
    lappend dirs [file dirname [info nameofexecutable]]\n\
    lappend dirs [file join [file dirname [file dirname [info nameofexecutable]]] lib]\n\
    lappend dirs [pwd]\n\
    if {[info exists env(TK_LIBRARY)]} {\n\
        lappend dirs $env(TK_LIBRARY)\n\
    }\n\
    lappend dirs $tk_library\n\
    lappend dirs [file join [file dirname [info library]] tk$tk_version]\n\
    lappend dirs [file join [file dirname [file dirname [info nameofexecutable]]] tk$tk_version]\n\
    if [string match {*[ab]*} $tk_patchLevel] {\n\
        set lib tk$tk_patchLevel\n\
    } else {\n\
        set lib tk$tk_version\n\
    }\n\
    lappend dirs [file join [file dirname [file dirname [pwd]]] $lib/library]\n\
    lappend dirs [file join [file dirname [pwd]] library]\n\
    lappend dirs ../../lib\n\
    foreach i $dirs {\n\
        set tkfile [file join $i tkxg.tcl]\n\
        if {[file exists $tkfile] && [file readable $tkfile]} {\n\
            set tcl_rcFileName $tkfile\n\
            return\n\
        }\n\
    }\n\
    set msg \"Can't find a usable tkxg.tcl in the following directories: \n\"\n\
    append msg \"    $dirs\n\n\"\n\
    append msg \"$errors\n\n\"\n\
    append msg \"This probably means that Tkxg wasn't installed properly.\n\"\n\
    error $msg\n\
}\n\
tkxgInit";

/*
 * startup call looking for tkxg.tcl and initializing globals as needed.
 */
int Tkxg_Init(interp)
Tcl_Interp *interp;
{
    static char st1[] = { 0x01 };
    Tk_DefineBitmap(interp,Tk_GetUid("apixel"),st1,1,1);

    /* hunt up rcfile. steal the tk style of doing it. */
    if (Tcl_Eval(interp,initScript)!=TCL_OK) {
        return TCL_ERROR;
    }
    init_tk(interp);
    /* define our the instance creator */
    Tcl_CreateCommand(interp,"tkxg",Tkxg_Create,NULL,NULL);
  
    tkxg_depth = Tk_Depth(Tk_MainWindow(interp));
    if (tkxg_depth >=4) {
        tkspecs = tkspecs_color;
    } else {
        tkspecs = tkspecs_bw;
    }
    return TCL_OK;
}

static
void do_error(m)
char *m;
{
    fprintf(stderr,"%s\n",m);
}
