/*
 * Xgraph Parameters
 *
 * This file contains routines for setting and retrieving
 * various X style display parameters for xgraph.
 */

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include "tcl.h"
#include "tk.h"
#include "st.h"
#include "params.h"

/* For use by convenience macros */
params param_temp, *param_temp_ptr;
XColor param_null_color = { 0, 0, 0, 0, 0, 0 };
param_style param_null_style = { STYLE, 0, (char *) 0 };

/*
 * this should be set to the interpreter of the canvas
 * being plotted. We should get to passing this around.
 */
Tcl_Interp *g_xgraph_interp;


typedef struct param_full_defn {
    param_types type;
    char *text_form;
    params *real_form;
} param_full;

static int tkxg_bitmapcount=0;

static void free_resource();
static params *resolve_entry();
static int strihash();

#define DEF_INT		"0"
#define DEF_STR		""
#ifdef __WIN32__
#define DEF_FONT	"system"
#else
#define DEF_FONT	"fixed"
#endif
#define DEF_PIXEL	"black"
#define DEF_STYLE	"1"
#define DEF_BOOL	"false"
#define DEF_DBL		"0.0"

#define DEF_MAX_FONT	1024
#define DEF_MAX_NAMES	10

#define DUP(str)	\
strcpy((char *) malloc((unsigned) (strlen(str)+1)), (str))


char *param_null_tkcolor = DEF_PIXEL;
char *param_null_tkfont = DEF_FONT;

st_table *param_init(interp)
Tcl_Interp *interp;
/*
 * Initializes parameter package. 
 */
{
    (void)interp;
    return st_init_table(stricmp, strihash);
}



void param_set(pt,name, type, val)
st_table *pt;
char *name;			/* Name of parameter   */
param_types type;		/* Type                */
char *val;			/* Text form for value */
/*
 * Sets the parameter with the given name to have the type
 * `type' and the text value `value'.  This will be evaluated
 * to its full form the first time it is referenced using
 * param_get().  If it is already filled, the old value
 * will be reclaimed.
 */
{
    param_full *entry;

    if (!pt) {
	(void) fprintf(stderr, "Parameter table not initialized\n");
	return;
    }
    if (st_lookup(pt, name, (char **) &entry)) {
	if (entry->real_form) free_resource(entry->real_form);
	entry->real_form = (params *) 0;
    } else {
	entry = (param_full *) malloc(sizeof(param_full));
	entry->text_form = (char *) 0;
	entry->real_form = (params *) 0;
	(void) st_insert(pt, DUP(name), (char *) entry);
    }
    entry->type = type;
    if (entry->text_form!=NULL) (void) free((char *) (entry->text_form));
    entry->text_form = DUP(val);
}

static 
enum st_retval param_delete(key,val,arg)
char *key;
param_full *val;
char *arg;
{
    (void)arg;
    if (key !=NULL) free(key); /* provide all keys from malloc'd strings */
    if (val->real_form != NULL) free_resource(val->real_form);
    val->real_form = (params *) 0;
    if (val->text_form!=NULL) (void) free((char *) (val->text_form));
    val->text_form = NULL;
    val->type = ERR;
    free(val);
    return ST_CONTINUE;
}

void param_destroy(pt)
st_table *pt;
/*
 * Sets the parameters in the given table to have NULL text forms
 * and releases any resources they may have.
 * The table should be destroyed immediately following this operation.
 */
{
    if (!pt) {
	(void) fprintf(stderr, "Parameter table not initialized\n");
	return;
    }
    st_foreach(pt, param_delete, (char *) 0);
    st_free_table(pt);
}


void param_reset(pt,name, val)
st_table *pt;
char *name;			/* Name of parameter   */
char *val;			/* Text form for value */
/*
 * This routine sets the value of an existing parameter to a new
 * value.  The type of the parameter remains the same.  Changes
 * in type should be done by using param_set() directly.
 */
{
    param_full *entry;

    if (!pt) {
	(void) fprintf(stderr, "Parameter table not initialized\n");
	return;
    }
    if (st_lookup(pt, name, (char **) &entry)) {
	param_set(pt,name, entry->type, val);
    } else {
	(void) fprintf(stderr, "Cannot reset unknown parameter `%s'\n", name);
    }
}




params *param_get(pt,name, val, interp)
st_table *pt;
char *name;			/* Name of parameter */
params *val;			/* Result value      */
Tcl_Interp *interp;
/*
 * Retrieves a value from the parameter table.  The value
 * is placed in `val'.  If successful, the routine will
 * return `val'.  Otherwise, it will return zero.
 */
{
    param_full *entry;

    if (!pt) {
	(void) fprintf(stderr, "Parameter table not initialized\n");
	return (params *) 0;
    }
    if (st_lookup(pt, name, (char **) &entry)) {
	if (!entry->real_form) {
	    entry->real_form = resolve_entry(name, entry->type,
                               entry->text_form, interp);
	}
	*val = *(entry->real_form);
	return val;
    } else {
	return (params *) 0;
    }
}


static void free_resource(val)
params *val;			/* Value to free */
/*
 * Reclaims a resource based on its type.
 */
{
    switch (val->type) {
    case INT:
    case STR:
    case BOOL:
    case DBL:
	/* No reclaimation necessary */
	break;
    case PIXEL:
	if (val->pixv.value!=NULL) {
            Tk_FreeColor(val->pixv.value);
            val->pixv.value = NULL;
        }
        free(val->pixv.tkcolorname);
        val->pixv.tkcolorname = NULL;
	break;
    case FONT:
        if (val->fontv.value != NULL) {
	    Tk_FreeFont(val->fontv.value);
	    val->fontv.value = NULL;
        }
        free(val->fontv.tkfontname);
        val->fontv.tkfontname = NULL;
	break;
    case STYLE:
	if (val->stylev.dash_list) {
            free(val->stylev.dash_list);
            val->stylev.dash_list = NULL;
        }
	if (val->stylev.tkbitname) {
            free(val->stylev.tkbitname);
            val->stylev.tkbitname = NULL;
        }
	/* do not free bitmap bits because we gave them to tk */
	break;
    default:
        break;
    }
    (void) free((char *) val);
}



static int do_color(interp,name,color)
Tcl_Interp *interp;
char *name;			/* Name for color */
XColor **color;
/*
 * Translates `name' into a Tk color.
 */
{
    int result = 1;
    Tk_Uid cid;
    cid = Tk_GetUid(name);
    *color = Tk_GetColor(interp,Tk_MainWindow(interp),name);
    return result;
}


static int do_font(interp, name, font_info)
Tcl_Interp *interp;
char *name;			/* Name of desired font      */
Tk_Font *font_info;	/* Returned font information */
/*
 * This routine translates a font name into a font structure. 
 * Any name acceptable by Tk_GetFont is allowable input.
 */
{
    *font_info = Tk_GetFont(interp,Tk_MainWindow(interp),name);
    if (*font_info==NULL) {
      return 0;
    }
    return 1;
}


static int do_style(list, val, interp)
char *list;			/* List of ones and zeros */
param_style *val;		/* Line style returned    */
Tcl_Interp *interp;
/*
 * Translates a string representation of a dash specification into
 * a form suitable for use in XSetDashes().  Assumes `list'
 * is a null terminated string of ones and zeros.
 * this is a holdover from unix where XDrawLine took a list of 01's
 * and did the right thing. the transformation into stipple here is
 * rather iffy, and many stipples may yield lines with a horizontal
 * bias.
 */
{
    char *i, *spot, last_char, *tkxbits;
    int count,c;
    unsigned bitlen;

    bitlen = 0;
    /* check of other than binary string value */
    for (i = list;  *i != '\0';  i++) {
	if ((*i != '0') && (*i != '1')) break;
        bitlen++;
    }
    if (!*i) { /* always TRUE for NULL terminated string unless garbage in */
        /* xgraph process */
	val->len = 0;
	last_char = '\0';
	for (i = list;  *i ;  i++) {
	    if (*i != last_char) {
		val->len += 1;
		last_char = *i;
	    }
	}
	val->dash_list = 
           (char *) malloc((unsigned) (sizeof(char)*val->len + 1));
	last_char = *list;
	spot = val->dash_list;
	count = 0;
	for (i = list;  *i;  i++) {
	    if (*i != last_char) {
		*spot++ = (char) count;
		last_char = *i;
		count = 1;
	    } else {
		count++;
	    }
	}
	*spot = (char) count;

        /* tk process */
        val->tkbitname = (char *) malloc(45); /* leaking? */
        sprintf(val->tkbitname,"%s%d",TKXGBITMAPROOT,tkxg_bitmapcount++); 
        /* build the NxN bitmap definition and register via tk */
        /* bitmap ultimately for stippling lines */
        tkxbits = calloc(1,bitlen+1); /* handed to tk/x. freeing can't be
                                     * done by them */
        for (c=0; c <bitlen;c++) {
          if (list[c]=='1') {
            tkxbits[c]= 0x1; 
          }
        }
        tkxbits[bitlen] = '\0'; /* perhaps not strictly necessary */
        Tk_DefineBitmap(interp,Tk_GetUid(val->tkbitname),
                        BITDATA(tkxbits),1,bitlen);
	return 1;
    } else {
	return 0;
    }
}


static char *positive[] = {
  "on", "yes", "true", "1", "affirmative", "si", "aye","y","Y",(char *) 0
};
static char *negative[] = { 
 "off", "no", "false", "0", "negative","nil","non","nein",
 "n","N","nope",(char *) 0 
};

static int do_bool(name, val)
char *name;			/* String representation */
int *val;			/* Returned value        */
/*
 * Translates a string representation into a suitable binary value.
 * Can parse all kinds of interesting boolean type words.
 */
{
    char **term;

    for (term = positive;  *term;  term++) {
	if (stricmp(name, *term) == 0) break;
    }
    if (*term) {
	*val = 1;
	return 1;
    }
    for (term = negative;  *term;  term++) {
	if (stricmp(name, *term) == 0) break;
    }
    if (*term) {
	*val = 0;
	return 1;
    }
    return 0;
}



static params *resolve_entry(name, type, form, interp)
char *name;			/* Name of item for errors */
param_types type;		/* What type of thing */
char *form;			/* Textual form       */
Tcl_Interp *interp;
/*
 * Allocates and returns an appropriate parameter structure
 * by translating `form' into its native type as given by `type'.
 * If it can't translate the given form, it will fall back onto
 * the default.
 */
{
    params *result = (params *) calloc(1,sizeof(params));
    char *tkstring;

    result->type = type;
    switch (type) {
    case INT:
	if (sscanf(form, "%d", &result->intv.value) != 1) {
	    (void) fprintf(stderr,
			   "Parameter %s: can't translate `%s' into",
                          name,form);
	    (void) fprintf(stderr, " an integer (defaulting to `%s')\n",
			   DEF_INT);
	    result->intv.value = atoi(DEF_INT);
	}
	break;
    case STR:
	result->strv.value = form;
	break;
    case PIXEL:
        tkstring = NULL;
	if (!do_color(interp,form,&(result->pixv.value))) {
	    (void) fprintf(stderr, 
                           "Parameter %s: can't translate `%s' into",
			   name, form);
	    (void) fprintf(stderr, 
                           " a color (defaulting to `%s')\n",
			   DEF_PIXEL);
	    (void) do_color(interp,DEF_PIXEL,&(result->pixv.value));
            tkstring = malloc(strlen(DEF_PIXEL)+1);
            sprintf(tkstring,"%s",DEF_PIXEL);
	}
        if (tkstring == NULL) {
            tkstring = malloc(strlen(form)+1);
            sprintf(tkstring,"%s",form);
        }
        result->pixv.tkcolorname = tkstring;
        tkstring = NULL;
	break;
    case FONT:
        tkstring = NULL;
	if (!do_font(interp,form, &result->fontv.value)) {
	    (void) fprintf(stderr,
                           "Parameter %s: can't translate `%s' into",
			   name, form);
	    (void) fprintf(stderr, " a font (defaulting to `%s')\n",
			   DEF_FONT);
	    (void) do_font(interp,DEF_FONT, &result->fontv.value);
            tkstring = malloc(strlen(DEF_FONT)+1);
            sprintf(tkstring,"%s",DEF_FONT);
	}
        if (tkstring == NULL) {
            tkstring = malloc(strlen(form)+1);
            sprintf(tkstring,"%s",form);
        }
        result->fontv.tkfontnamelen = strlen(tkstring);
        result->fontv.tkfontname = tkstring;
        tkstring = NULL;
	break;
    case STYLE:
	if (!do_style(form, &result->stylev,interp)) {
	    (void) fprintf(stderr,
               "Parameter %s: can't translate `%s' into a line style (defaulting to `%s')\n",
			   name, form, DEF_STYLE);
	    (void) do_style(DEF_STYLE, &result->stylev,interp);
	}
	break;
    case BOOL:
	if (!do_bool(form, &result->boolv.value)) {
	    (void) fprintf(stderr,
               "Parameter %s: can't translate `%s' into a boolean flag (defaulting to `%s')\n",
			   name, form, DEF_BOOL);
	    (void) do_bool(DEF_BOOL, &result->boolv.value);
	}
	break;
    case DBL:
	if (sscanf(form, "%lf", &result->dblv.value) != 1) {
	    (void) fprintf(stderr,
	       "Parameter %s: can't translate `%s' into a double (defaulting to `%s')\n",
			   name, form, DEF_DBL);
	    result->dblv.value = atof(DEF_DBL);
	}
	break;
    default:
         (void) fprintf(stderr,
              "Parameter %s: can't translate `%s' into an ERR parameter\n",
               name, form);
        
    }
    return result;
}




/*ARGSUSED*/
static enum st_retval dump_it(key, value, arg)
char *key, *value, *arg;
{
    param_full *val = (param_full *) value;

    (void)arg;
    printf("%s (", key);
    switch (val->type) {
    case INT:
	printf("INT");
	break;
    case STR:
	printf("STR");
	break;
    case PIXEL:
	printf("PIXEL");
	break;
    case FONT:
	printf("FONT");
	break;
    case STYLE:
	printf("STYLE");
	break;
    case BOOL:
	printf("BOOL");
	break;
    case DBL:
	printf("DBL");
	break;
    default:
	printf("ERR");
	break;
    }
    printf(") = %s\n", val->text_form);
    return ST_CONTINUE;
}

void param_dump(pt)
st_table *pt;
/*
 * Dumps all of the parameter values to standard output.
 */
{
    st_foreach(pt, dump_it, (char *) 0);
}


#ifndef __WIN32__
int stricmp(a, b)
register char *a, *b;
/*
 * This routine compares two strings disregarding case.
 */
{
    register int value;

    if ((a == (char *) 0) || (b == (char *) 0)) {
	return a - b;
    }

    value = 0;
    for ( /* nothing */;
	 ((*a | *b) &&
	  !(value = ((isupper(*a) ? *a - 'A' + 'a' : *a) -
		     (isupper(*b) ? *b - 'A' + 'a' : *b))));
	 a++, b++)
      /* Empty Body */;

    return value;
}
#endif /* win32 */

static int strihash(string, modulus)
register char *string;
int modulus;
/* Case insensitive computation */
{
    register int val = 0;
    register int c;
    
    while ((c = *string++) != '\0') {
	if (isupper(c)) c = tolower(c);
	val = val*997 + c;
    }

    return ((val < 0) ? -val : val)%modulus;
}
